//
//  certcheck.c
//  caqe-qcir
//
//  Copyright (c) 2016, Leander Tentrup, Saarland University
//
//  Licensed under Apache License v2.0, see LICENSE.txt for information
//

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "aiger.h"
#include "getopt.h"
#include "logging.h"
#include "map.h"
#include "matrix.h"
#include "qdimacs.h"


typedef struct {
    Matrix* matrix;
    qbf_res result;
    aiger* cert;
    aiger* combined;
    unsigned next_aiger_lit;
    
    int_queue clauses;
    int_queue clause_literals;
} CertCheck;

static qbf_res certificate_get_type(aiger* certificate) {
    // the last output of certificate encodes whether its a Skolem or Herbrand certificate
    aiger_symbol strategy_type = certificate->outputs[certificate->num_outputs - 1];
    if (strategy_type.lit == 0) {
        return QBF_RESULT_UNSAT;
    } else {
        assert(strategy_type.lit == 1);
        return QBF_RESULT_SAT;
    }
}

static CertCheck* certcheck_init(Matrix* matrix, aiger* cert) {
    CertCheck* check = malloc(sizeof(CertCheck));
    check->matrix = matrix;
    check->next_aiger_lit = matrix->max_vars + 1;
    check->result = certificate_get_type(cert);
    check->cert = cert;
    check->combined = aiger_init();
    int_queue_init(&check->clauses);
    int_queue_init(&check->clause_literals);
    return check;
}

/**
 * Translates matrix literals to AIGER literals.
 *
 * Example: var = -5 => return 5 * 2 + 1
 */
static unsigned matrix_var_to_aiger_lit(int var) {
    if (var < 0) {
        var = -var;
        return (unsigned)var * 2 + 1;
    }
    return (unsigned)var * 2;
}

static char* int2str(int i) {
    size_t buffer_size = (((size_t)ceil(log10(i)) + 1) * sizeof(char));
    char* str = malloc(buffer_size);
    sprintf(str, "%d", i);
    return str;
}

static unsigned new_aiger_lit(CertCheck* check) {
    return ++check->next_aiger_lit * 2;
}

/**
 * Encodes the AND of aiger literals given in int_vector such that the result
 * is stored in dest_aiger_lit
 *
 * @note modifies aiger_lits
 */
static void encode_and_as(CertCheck* check, int_queue* aiger_lits, unsigned dest_aiger_lit) {
    while (int_queue_has_at_least_two_elements(aiger_lits)) {
        unsigned lhs = (unsigned)int_queue_pop(aiger_lits);
        unsigned rhs = (unsigned)int_queue_pop(aiger_lits);
        unsigned new_lit = new_aiger_lit(check);
        aiger_add_and(check->combined, new_lit, lhs, rhs);
        int_queue_push(aiger_lits, (int)new_lit);
    }
    unsigned lhs = aiger_true, rhs = aiger_true;
    
    if (int_queue_has_at_least_one_element(aiger_lits)) {
        lhs = (unsigned)int_queue_pop(aiger_lits);
    }
    if (int_queue_has_at_least_one_element(aiger_lits)) {
        rhs = (unsigned)int_queue_pop(aiger_lits);
    }
    aiger_add_and(check->combined, aiger_strip(dest_aiger_lit), lhs, rhs);
}
/**
 * Encodes the OR of aiger literals given in int_vector such that the result
 * is stored in dest_aiger_lit
 *
 * @note modifies aiger_lits
 */
static void encode_or_as(CertCheck* check, int_queue* aiger_lits, unsigned dest_aiger_lit) {
    // negate literals in aiger_lits first
    for (int_queue_element* element = aiger_lits->first; element != NULL; element = element->next) {
        element->value = (int)aiger_not(element->value);
    }
    encode_and_as(check, aiger_lits, dest_aiger_lit);
}

static void import_variables(CertCheck* check) {
    quant_t func_type = (check->result == QBF_RESULT_SAT) ? QUANTIFIER_EXISTENTIAL : QUANTIFIER_UNIVERSAL;
    for (size_t i = 1; i <= check->matrix->max_vars; i++) {
        Var* var = check->matrix->vars[i];
        if (var == NULL) {
            continue;
        }
        if (var->scope->qtype == func_type) {
            continue;
        }
        aiger_add_input(check->combined, matrix_var_to_aiger_lit(var->var_id), int2str(var->var_id));
    }
}

static int import_clause(CertCheck* check, Clause* clause) {
    assert(int_queue_is_empty(&check->clause_literals));
    
    for (Occ* occ = clause->first_occ; occ != NULL; occ = occ->next_in_clause) {
        const int literal = matrix_get_lit(occ);
        unsigned aiger_literal = matrix_var_to_aiger_lit(literal);
        if (check->result == QBF_RESULT_UNSAT) {
            aiger_literal = aiger_not(aiger_literal);
        }
        int_queue_push(&check->clause_literals, aiger_literal);
    }
    
    int clause_result_literal = new_aiger_lit(check);
    if (check->result == QBF_RESULT_SAT) {
        encode_or_as(check, &check->clause_literals, clause_result_literal);
        return aiger_not(clause_result_literal);
    } else {
        assert(check->result == QBF_RESULT_UNSAT);
        encode_and_as(check, &check->clause_literals, clause_result_literal);
        return clause_result_literal;
    }
}

static void import_cnf(CertCheck* check) {
    const Matrix* matrix = check->matrix;
    
    assert(int_queue_is_empty(&check->clauses));
    
    for (Clause* clause = matrix->first_clause; clause != NULL; clause = clause->next) {
        int_queue_push(&check->clauses, import_clause(check, clause));
    }
    
    int clauses_result_literal = new_aiger_lit(check);
    if (check->result == QBF_RESULT_SAT) {
        encode_and_as(check, &check->clauses, clauses_result_literal);
        aiger_add_output(check->combined, aiger_not(clauses_result_literal), "bad");
    } else {
        assert(check->result == QBF_RESULT_UNSAT);
        encode_or_as(check, &check->clauses, clauses_result_literal);
        aiger_add_output(check->combined, clauses_result_literal, "bad");
    }
}

static unsigned translate_src_input_to_dst_input(aiger_symbol* input, aiger* dst) {
    // need to do linear search through inputs
    for (size_t i = 0; i < dst->num_inputs; i++) {
        aiger_symbol cmp = dst->inputs[i];
        if (strcmp(cmp.name, input->name) == 0) {
            return cmp.lit;
        }
    }
    assert(false);
    return 0;
}

/**
 * Translates a AIGER literal from src to destination.
 *
 * - Constants are not touched
 * - Inputs are translated using translate_src_input_to_dst_input
 * - ANDs are translated using the given offset
 */
static unsigned aiger_lit_translate(aiger* src, aiger* dest, unsigned lit, unsigned offset) {
    aiger_symbol* input = NULL;
    if (aiger_strip(lit) == 0) {
        // constant
        return lit;
    } else if ((input = aiger_is_input(src, aiger_strip(lit)))) {
        // input
        int translated = translate_src_input_to_dst_input(input, dest);
        return (aiger_strip(lit) == lit) ? translated : aiger_not(translated);
    } else {
        return lit + offset;
    }
}

static void import_and_gates(aiger* src, aiger* dest, unsigned offset) {
    for (unsigned i = 0; i < src->num_ands; i++) {
        aiger_and and_gate = src->ands[i];
        
        unsigned shifted_and = and_gate.lhs + offset;
        unsigned shifted_lhs = aiger_lit_translate(src, dest, and_gate.rhs0, offset);
        unsigned shifted_rhs = aiger_lit_translate(src, dest, and_gate.rhs1, offset);
        
        aiger_add_and(dest, shifted_and, shifted_lhs, shifted_rhs);
    }
}

static void define_strategies(CertCheck* check) {
    // the last output indicates whether certificate is Skolem/Herbrand so we ignore it here
    for (unsigned i = 0; i < check->cert->num_outputs - 1; i++) {
        aiger_symbol output = check->cert->outputs[i];
        unsigned orig_var = strtol(output.name, NULL, 0);
        assert(orig_var > 0);
        assert(orig_var <= check->matrix->max_vars);
        assert(check->matrix->vars[orig_var] != NULL);
        
        unsigned output_definition = aiger_lit_translate(check->cert, check->combined, output.lit, check->next_aiger_lit * 2);
        aiger_add_and(check->combined, matrix_var_to_aiger_lit(orig_var), output_definition, 1);
    }
}

/**
 * Imports circuit representing strategy into check->combined.
 * - Checks that inputs of both circuits are the same
 * - Offsets all AND gates by check->next_aig_lit
 * - Defines AND gates representing strategy variables
 */
static void combine_with_certificate(CertCheck* check) {
    import_and_gates(check->cert, check->combined, check->next_aiger_lit * 2);
    define_strategies(check);
}


/**
 * Convert the CNF matrix to AIGER
 */
static void import_matrix(CertCheck* check) {
    import_variables(check);
    import_cnf(check);
}


static void print_usage(const char* name) {
    printf("usage: %s [-h] instance cert\n\n"
           "optional arguments:\n"
           "  -h, --help\t show this help message and exit\n", name);
}

int main(int argc, char * const argv[]) {
    // Handling of command line arguments
    opterr = 0; // Disable error messages as we handle them ourself
    const char * ch;
    volatile bool print_size = false;
    while ((ch = GETOPT(argc, argv)) != NULL) {
        GETOPT_SWITCH(ch) {
            GETOPT_OPT("-h"):
            print_usage(argv[0]);
            return 1;
            break;
            GETOPT_OPT("-s"):
            print_size = true;
            break;
        GETOPT_MISSING_ARG:
            printf("missing argument to %s\n", ch);
            /* FALLTHROUGH */
        GETOPT_DEFAULT:
            printf("%s", ch);
            print_usage(argv[0]);
            return 1;
        }
    }
    
    if (optind + 2 < argc) {
        logging_error("too many arguments, see --help for more information\n");
        return 1;
    } else if (optind + 1 > argc) {
        logging_error("too few arguments, see --help for more information\n");
        return 1;
    }
    
    const char* instance_name = argv[optind];
    bool use_stdin = optind + 2 > argc;
    const char* cert_name;
    if (!use_stdin) {
        cert_name = argv[optind + 1];
    }
    
    Matrix* matrix = matrix_init();
    if (!matrix) {
        return 1;
    }
    
    int error = qdimacs_open_and_read_qcir_file(matrix, instance_name);
    if (error) {
        return 1;
    }
    
    aiger* cert = aiger_init();
    if (cert == NULL) {
        return 1;
    }
    
    if (use_stdin) {
        if (aiger_read_from_file(cert, stdin)) {
            logging_fatal("cannot read certificate file\n");
            return  1;
        }
    } else {
        if (aiger_open_and_read_from_file(cert, cert_name) != NULL) {
            logging_fatal("cannot read certificate file\n");
            return  1;
        }
    }
    
    if (print_size) {
        fprintf(stderr, "Size: %d\n", cert->num_ands);
    }
    
    CertCheck* check = certcheck_init(matrix, cert);
    import_matrix(check);
    combine_with_certificate(check);
    
    if (aiger_check(check->combined)) {
        printf("%s\n", aiger_error(check->combined));
        abort();
    }
    
    assert(!aiger_check(check->combined));

    aiger_reencode(check->combined);
    aiger_write_to_file(check->combined, aiger_ascii_mode, stdout);
    
    matrix_free(matrix);
    
    return 0;
}

