//
//  certification.c
//  caqe-qcir
//
//  Copyright (c) 2016, Leander Tentrup, Saarland University
//
//  Licensed under Apache License v2.0, see LICENSE.txt for information
//

#include "certification.h"

#include <assert.h>

#include "aiger_helper.h"
#include "clausal_abstraction.h"

static char buffer[1024] = { 0 };

static const char* int2str(unsigned literal) {
    snprintf(buffer, 1024, "%d", literal);
    return buffer;
}

void certification_init(certification* cert, Matrix* matrix) {
    cert->skolem = aiger_init();
    cert->herbrand = aiger_init();
    cert->next_lit_skolem = matrix->max_vars + 1 + (matrix->last_scope_id + 1) * matrix->num_clauses;
    cert->next_lit_herbrand = matrix->max_vars + 1 + (matrix->last_scope_id + 1) * matrix->num_clauses;
    
    int_queue_init(&cert->queue);
    
    cert->function_lit = map_init();
    cert->precondition_lit = map_init();
}

static unsigned matrix_lit_to_aiger_lit(lit_t literal) {
    if (literal < 0) {
        literal = -literal;
        return aiger_not((unsigned)literal * 2);
    }
    return (unsigned)literal * 2;
}

unsigned certification_define_and(certification*, quant_t);

static void queue_push(certification* cert, unsigned value) {
    if (value == 1) {
        return;
    }
    int_queue_push(&cert->queue, (lit_t)value);
}

void certification_import_variables(certification* cert, Matrix* matrix) {
    for (size_t i = 1; i <= matrix->max_vars; i++) {
        Var* var = matrix->vars[i];
        if (var == NULL) {
            continue;
        }
        unsigned value = aiger_false;
        unsigned precondition = aiger_false;

        while (!int_queue_is_empty(&var->reductions)) {
            // reductions have the form (value literals 0)..(value literals 0)
            int var_value = int_queue_pop(&var->reductions);
            assert(var_value == 1 || var_value == -1);
            lit_t literal;
            assert(!int_queue_is_empty(&var->reductions));
            assert(certification_queue_is_empty(cert));
            while ((literal = int_queue_pop(&var->reductions))) {
                certification_add_literal(cert, -literal);
            }
            unsigned current_precondition = certification_define_and(cert, QUANTIFIER_UNIVERSAL);
            if (var_value > 0) {
                // !last_precondition && precondition_aiger_lit => x
                assert(int_queue_is_empty(&cert->queue));
                queue_push(cert, aiger_not(precondition));
                queue_push(cert, current_precondition);
                unsigned function_case = certification_define_and(cert, QUANTIFIER_UNIVERSAL);
                
                // function_lit |= function_case
                assert(int_queue_is_empty(&cert->queue));
                queue_push(cert, aiger_not(value));
                queue_push(cert, aiger_not(function_case));
                value = aiger_not(certification_define_and(cert, QUANTIFIER_UNIVERSAL));
            }
            
            // last_precondition |= precondition_aiger_lit
            assert(int_queue_is_empty(&cert->queue));
            queue_push(cert, aiger_not(precondition));
            queue_push(cert, aiger_not(current_precondition));
            precondition = aiger_not(certification_define_and(cert, QUANTIFIER_UNIVERSAL));
        }
        if (var->removed && var->value > 0) {
            // function_lit |= function_case
            assert(int_queue_is_empty(&cert->queue));
            queue_push(cert, aiger_not(value));
            queue_push(cert, precondition);
            value = aiger_not(certification_define_and(cert, QUANTIFIER_UNIVERSAL));
        }
        if (var->removed || matrix->conflict) {
            if (var->orig_quant == QUANTIFIER_EXISTENTIAL) {
                aiger_add_output(cert->skolem, value, int2str(var->orig_var_id));
                aiger_add_input(cert->herbrand, var->var_id * 2, int2str(var->orig_var_id));
            } else {
                aiger_add_output(cert->herbrand, value, int2str(var->orig_var_id));
                aiger_add_input(cert->skolem, var->var_id * 2, int2str(var->orig_var_id));
            }
        } else {
            assert(var->scope != NULL);
            assert(var->orig_quant == var->scope->qtype);
            if (var->orig_quant == QUANTIFIER_EXISTENTIAL) {
                aiger_add_output(cert->skolem, var->var_id * 2, int2str(var->orig_var_id));
                aiger_add_input(cert->herbrand, var->var_id * 2, int2str(var->orig_var_id));
            } else {
                assert(var->orig_quant == QUANTIFIER_UNIVERSAL);
                aiger_add_output(cert->herbrand, var->var_id * 2, int2str(var->orig_var_id));
                aiger_add_input(cert->skolem, var->var_id * 2, int2str(var->orig_var_id));
            }
            
            map_add(cert->function_lit, var->var_id, (void*)(uintptr_t)value);
            map_add(cert->precondition_lit, var->var_id, (void*)(uintptr_t)value);
        }
    }
}

void certification_add_literal(certification* cert, lit_t literal) {
    int_queue_push(&cert->queue, matrix_lit_to_aiger_lit(literal));
}

void certification_add_aiger_literal(certification* cert, lit_t aiger_lit) {
    int_queue_push(&cert->queue, aiger_lit);
}

static unsigned translate_clause_to_aiger_literal(ClausalAbstraction* abs, quant_t qtype, var_t clause_id) {
    // scope_id can be zero
    const unsigned offset = abs->solver->matrix->max_vars + 1 + abs->scope->scope_id * abs->solver->matrix->num_clauses;
    const unsigned literal = offset + clause_id;
    unsigned aiger_literal = literal * 2;
    
    if (qtype == QUANTIFIER_EXISTENTIAL) {
        aiger_literal = aiger_not(aiger_literal);
    }
    return aiger_literal;
}

/*void certification_add_b_literal(certification* cert, CircuitAbstraction* abs, quantifier_type qtype, var_t gate_id) {
    unsigned translated_b_lit = translate_gate_to_b_literal(abs, qtype, gate_id);
    int_queue_push(&cert->queue, translated_b_lit);
}*/

void certification_add_t_literal(certification* cert, ClausalAbstraction* abs, quant_t qtype, var_t clause_id, lit_t b_lit, lit_t t_lit) {
    // traverse scopes until we hit an abstraction that has a b-literal corresponding to gate_id
    for (ClausalAbstraction* outer = abs->prev; outer != NULL; outer = outer->prev) {
        if (!int_vector_contains_sorted(outer->b_lits, b_lit) && !int_vector_contains_sorted(outer->b_lits, t_lit)) {
            assert(outer->prev != NULL);
            continue;
        }
        int_queue_push(&cert->queue, translate_clause_to_aiger_literal(outer, qtype, clause_id));
        return;
    }
}

void certification_define_b_literal(certification* cert, ClausalAbstraction* abs, quant_t qtype, var_t clause_id) {
    aiger* base = (qtype == QUANTIFIER_EXISTENTIAL) ? cert->skolem : cert->herbrand;
    unsigned* next_lit = (qtype == QUANTIFIER_EXISTENTIAL) ? &cert->next_lit_skolem : &cert->next_lit_herbrand;
    
    const unsigned translated_b_lit = translate_clause_to_aiger_literal(abs, qtype, clause_id);
    if (qtype == QUANTIFIER_EXISTENTIAL) {
        encode_or_as(base, next_lit, &cert->queue, aiger_strip(translated_b_lit));
    } else {
        assert(qtype == QUANTIFIER_UNIVERSAL);
        encode_and_as(base, next_lit, &cert->queue, translated_b_lit);
    }
}

unsigned certification_define_or(certification* cert, quant_t qtype) {
    aiger* base = (qtype == QUANTIFIER_EXISTENTIAL) ? cert->skolem : cert->herbrand;
    unsigned* next_lit = (qtype == QUANTIFIER_EXISTENTIAL) ? &cert->next_lit_skolem : &cert->next_lit_herbrand;
    
    return encode_or(base, next_lit, &cert->queue);
}

unsigned certification_define_and(certification* cert, quant_t qtype) {
    aiger* base = (qtype == QUANTIFIER_EXISTENTIAL) ? cert->skolem : cert->herbrand;
    unsigned* next_lit = (qtype == QUANTIFIER_EXISTENTIAL) ? &cert->next_lit_skolem : &cert->next_lit_herbrand;
    
    return encode_and(base, next_lit, &cert->queue);
}

/*void certification_push(certification* cert) {
    cert->current_queue++;
    assert(cert->current_queue < 2);
}

void certification_pop(certification* cert) {
    assert(cert->current_queue > 0);
    cert->current_queue--;
}

void certification_clear(certification* cert) {
    while (!int_queue_is_empty(&cert->queue)) {
        int_queue_pop(&cert->queue);
    }
}*/

bool certification_queue_is_empty(certification* cert) {
    return int_queue_is_empty(&cert->queue);
}

void certification_append_function_case(certification* cert, ClausalAbstraction* abs) {
    unsigned current_precondition = certification_define_and(cert, abs->scope->qtype);
    for (size_t i = 0; i < vector_count(abs->scope->vars); i++) {
        const Var* var = vector_get(abs->scope->vars, i);
        bool negated = (var->value <= 0);
        
        unsigned function_lit = (uintptr_t)map_get(cert->function_lit, var->var_id);
        unsigned last_precondition = (uintptr_t)map_get(cert->precondition_lit, var->var_id);
        
        if (!negated) {
            // !last_precondition && precondition_aiger_lit => x
            assert(int_queue_is_empty(&cert->queue));
            queue_push(cert, aiger_not(last_precondition));
            queue_push(cert, current_precondition);
            unsigned function_case = certification_define_and(cert, abs->scope->qtype);
            
            // function_lit |= function_case
            assert(int_queue_is_empty(&cert->queue));
            queue_push(cert, aiger_not(function_lit));
            queue_push(cert, aiger_not(function_case));
            unsigned new_function_lit = certification_define_and(cert, abs->scope->qtype);
            map_update(cert->function_lit, var->var_id, (void*)(uintptr_t)aiger_not(new_function_lit));
        }
        
        // last_precondition |= precondition_aiger_lit
        assert(int_queue_is_empty(&cert->queue));
        queue_push(cert, aiger_not(last_precondition));
        queue_push(cert, aiger_not(current_precondition));
        unsigned new_precondition = certification_define_and(cert, abs->scope->qtype);
        
        map_update(cert->precondition_lit, var->var_id, (void*)(uintptr_t)aiger_not(new_precondition));
    }
}

void certification_define_outputs(certification* cert, ClausalAbstraction* abs, qbf_res result) {
    aiger* strategy = (abs->scope->qtype == QUANTIFIER_EXISTENTIAL) ? cert->skolem : cert->herbrand;
    if ((abs->scope->qtype == QUANTIFIER_EXISTENTIAL && result == QBF_RESULT_SAT)
        || (abs->scope->qtype == QUANTIFIER_UNIVERSAL && result == QBF_RESULT_UNSAT)) {
        
        for (size_t i = 0; i < vector_count(abs->scope->vars); i++) {
            Var* var = vector_get(abs->scope->vars, i);
            
            // define output as AND gate such that outer variables can depend on it
            unsigned function_lit = (uintptr_t)map_get(cert->function_lit, var->var_id);
            aiger_add_and(strategy, var->var_id * 2, function_lit, 1);
        }
    }
    
    for (size_t i = 0; i < abs->scope->num_next; i++) {
        certification_define_outputs(cert, abs->next[i], result);
    }
}

void certification_print(certification* cert, qbf_res result) {
    aiger* strategy = (result == QBF_RESULT_SAT) ? cert->skolem : cert->herbrand;
    if (aiger_check(strategy)) {
        printf("%s\n", aiger_error(strategy));
        abort();
    }
    unsigned result_output = (result == QBF_RESULT_SAT) ? aiger_true : aiger_false;
    aiger_add_output(strategy, result_output, "result");
    if (result == QBF_RESULT_SAT) {
        aiger_add_comment(strategy, "SAT");
    } else {
        assert(result == QBF_RESULT_UNSAT);
        aiger_add_comment(strategy, "UNSAT");
    }
    aiger_reencode(strategy);
    aiger_write_to_file(strategy, aiger_ascii_mode, stdout);
}
