//
//  clausal_abstraction.c
//  caqe
//
//  Copyright (c) 2015 - 2016, Leander Tentrup, Saarland University
//                2015, Markus Rabe, University of California, Berkeley
//
//  Licensed under Apache License v2.0, see LICENSE.txt for information
//

#include "clausal_abstraction.h"

#include <assert.h>
#include <stdint.h>

#include "certification.h"
#include "logging.h"
#include "solver.h"

int clause_to_t_lit(Solver* solver, Clause* clause) {
    return clause->clause_id + solver->matrix->max_vars + 1;
}

int clause_to_b_lit(Solver* solver, Clause* clause) {
    return t_lit_to_b_lit(solver, clause_to_t_lit(solver, clause));
}

int b_lit_to_t_lit(Solver* solver, int b_lit) {
    return b_lit - (solver->matrix->num_clauses);
}

int b_lit_to_clause_id(Solver* solver, int b_lit) {
    return t_lit_to_clause_id(solver, b_lit_to_t_lit(solver, b_lit));
}

int t_lit_to_b_lit(Solver* solver, int t_lit) {
    return t_lit + solver->matrix->num_clauses;
}

int t_lit_to_clause_id(Solver* solver, int t_lit) {
    return t_lit - (solver->matrix->max_vars + 1);
}

static void build_existential_abstraction(Solver* solver, ClausalAbstraction* abstraction, bool primary) {
    const Scope* scope = abstraction->scope;
    const Matrix* matrix = solver->matrix;
    SATSolver* sat = primary ? abstraction->sat : abstraction->negation;
    
    // build SAT instance
    for (Clause* clause = matrix_iterator_start(matrix, scope); clause != matrix_iterator_end(matrix, scope); clause = clause->next) {
        assert(clause->size != 0);  // unit clauses are handled outside
#ifdef CERTIFICATION
        assert(!solver->options->certify || certification_queue_is_empty(&solver->certificate));
#endif
        bool contains_variables = false;
        if (bit_vector_min(clause->scopes) > scope->scope_id) {
            continue;
        }
        for (Occ* occ = clause->first_occ; occ != NULL; occ = occ->next_in_clause) {
            if (occ->var->scope != scope) {
                continue;
            }
            satsolver_add(sat, matrix_get_lit(occ));
#ifdef CERTIFICATION
            if (solver->options->certify) {
                certification_add_literal(&solver->certificate, matrix_get_lit(occ));
            }
#endif
            contains_variables = true;
        }
        
        // add t- and b-lits to existential quantifiers:
        // * we add t-lit if scope is between min- and max-scope of current clause
        // * we add b-lit if scope is between min- and max-scope of current clause, excluding max-scope
        const bool need_t_lit = bit_vector_min(clause->scopes) < scope->scope_id && scope->scope_id <= bit_vector_max(clause->scopes);
        const bool need_b_lit = bit_vector_min(clause->scopes) <= scope->scope_id && scope->scope_id < bit_vector_max(clause->scopes);
        
        if (!need_t_lit && !contains_variables && !need_b_lit) {
            continue;
        }
        
        if (need_t_lit) {
            const int t_lit = clause_to_t_lit(solver, clause);
            satsolver_add(sat, t_lit);
            if (primary) {
                int_vector_add(abstraction->t_lits, t_lit);
            }
#ifdef CERTIFICATION
            if (solver->options->certify) {
                const int b_lit = clause_to_b_lit(solver, clause);
                certification_add_t_literal(&solver->certificate, abstraction, QUANTIFIER_EXISTENTIAL, clause->clause_id, b_lit, t_lit);
            }
#endif
        }
        if (need_b_lit) {
            const int b_lit = clause_to_b_lit(solver, clause);
            satsolver_add(sat, -b_lit);
            if (primary) {
                int_vector_add(abstraction->b_lits, b_lit);
            }
        }
        satsolver_add(sat, 0);
        
        if (scope->num_next != 0 && bit_vector_max(clause->scopes) == scope->scope_id) {
            assert(abstraction->local_refinement_literals != NULL);
            const int b_lit = clause_to_b_lit(solver, clause);
            int_vector_add(abstraction->local_refinement_literals, b_lit);
        }
        
#ifdef CERTIFICATION
        if (solver->options->certify) {
            certification_define_b_literal(&solver->certificate, abstraction, QUANTIFIER_EXISTENTIAL, clause->clause_id);
        }
#endif
    }
    
    if (logging_get_verbosity() >= VERBOSITY_ALL) {
        logging_debug("\n");
        logging_debug("level %d\n", scope->scope_id);
        satsolver_print(sat);
        
        int_vector_print(abstraction->t_lits);
        int_vector_print(abstraction->b_lits);
    }
    
    //assert(satsolver_sat(abstraction->sat) == SATSOLVER_SATISFIABLE);
}

static void build_universal_abstraction(Solver* solver, ClausalAbstraction* abstraction) {
    const Scope* scope = abstraction->scope;
    const Matrix* matrix = solver->matrix;
    SATSolver* sat = abstraction->sat;
    
    // build SAT instance
    for (Clause* clause = matrix_iterator_start(matrix, scope); clause != matrix_iterator_end(matrix, scope); clause = clause->next) {
        const int t_lit = clause_to_t_lit(solver, clause);
        if (bit_vector_min(clause->scopes) > scope->scope_id) {
            continue;
        }
        for (Occ* occ = clause->first_occ; occ != NULL; occ = occ->next_in_clause) {
            if (occ->var->scope != scope) {
                continue;
            }
            satsolver_add(sat, -t_lit);
            satsolver_add(sat, -matrix_get_lit(occ));
            satsolver_add(sat, 0);
            
#ifdef CERTIFICATION
            if (solver->options->certify) {
                certification_add_literal(&solver->certificate, -matrix_get_lit(occ));
            }
#endif
        }
        const bool need_t_lit = bit_vector_min(clause->scopes) < scope->scope_id && scope->scope_id <= bit_vector_max(clause->scopes);
        const bool need_b_lit = bit_vector_min(clause->scopes) <= scope->scope_id && scope->scope_id <= bit_vector_max(clause->scopes);
        if (need_t_lit) {
            int_vector_add(abstraction->t_lits, t_lit);
            
#ifdef CERTIFICATION
            if (solver->options->certify) {
                const int b_lit = clause_to_b_lit(solver, clause);
                certification_add_t_literal(&solver->certificate, abstraction, QUANTIFIER_UNIVERSAL, clause->clause_id, b_lit, t_lit);
            }
#endif
        }
        if (need_b_lit) {
            int_vector_add(abstraction->b_lits, t_lit);
        }
        
#ifdef CERTIFICATION
        if (solver->options->certify) {
            certification_define_b_literal(&solver->certificate, abstraction, QUANTIFIER_UNIVERSAL, clause->clause_id);
        }
#endif
    }
    
    if (logging_get_verbosity() >= VERBOSITY_ALL) {
        logging_debug("\n");
        logging_debug("level %d\n", scope->scope_id);
        satsolver_print(sat);
        
        int_vector_print(abstraction->t_lits);
        int_vector_print(abstraction->b_lits);
    }
    
    //assert(satsolver_sat(abstraction->sat) == SATSOLVER_SATISFIABLE);
}

static void build_dual_universal_abstraction(Solver* solver, ClausalAbstraction* abstraction) {
    const Scope* scope = abstraction->scope;
    const Matrix* matrix = solver->matrix;
    SATSolver* sat = abstraction->negation;
    
    // build SAT instance
    for (Clause* clause = matrix_iterator_start(matrix, scope); clause != matrix_iterator_end(matrix, scope); clause = clause->next) {
        const int t_lit = clause_to_t_lit(solver, clause);
        const int b_lit = clause_to_b_lit(solver, clause);
        if (bit_vector_min(clause->scopes) > scope->scope_id) {
            continue;
        }
        for (Occ* occ = clause->first_occ; occ != NULL; occ = occ->next_in_clause) {
            if (occ->var->scope != scope) {
                continue;
            }
            satsolver_add(sat, -b_lit);
            satsolver_add(sat, -matrix_get_lit(occ));
            satsolver_add(sat, 0);
            
#ifdef CERTIFICATION
            if (solver->options->certify) {
                certification_add_literal(&solver->certificate, -matrix_get_lit(occ));
            }
#endif
        }
        const bool need_t_lit = bit_vector_min(clause->scopes) < scope->scope_id && scope->scope_id <= bit_vector_max(clause->scopes);
        //const bool need_b_lit = bit_vector_min(clause->scopes) <= scope->scope_id && scope->scope_id <= bit_vector_max(clause->scopes);
        if (need_t_lit) {
            satsolver_add(sat, t_lit);
            satsolver_add(sat, -b_lit);
            satsolver_add(sat, 0);
            
#ifdef CERTIFICATION
            if (solver->options->certify) {
                certification_add_t_literal(&solver->certificate, abstraction, QUANTIFIER_UNIVERSAL, clause->clause_id, b_lit, t_lit);
            }
#endif
        }
        
#ifdef CERTIFICATION
        if (solver->options->certify) {
            certification_define_b_literal(&solver->certificate, abstraction, QUANTIFIER_UNIVERSAL, clause->clause_id);
        }
#endif
    }
    
    for (Clause* clause = matrix_iterator_start(matrix, scope); clause != matrix_iterator_end(matrix, scope); clause = clause->next) {
        const int b_lit = clause_to_b_lit(solver, clause);
        const bool need_b_lit = bit_vector_min(clause->scopes) <= scope->scope_id && scope->scope_id <= bit_vector_max(clause->scopes);
        if (need_b_lit) {
            satsolver_add(sat, b_lit);
        }
    }
    satsolver_add(sat, 0);
    
    if (logging_get_verbosity() >= VERBOSITY_ALL) {
        logging_debug("\n");
        logging_debug("level %d\n", scope->scope_id);
        satsolver_print(sat);
        
        int_vector_print(abstraction->t_lits);
        int_vector_print(abstraction->b_lits);
    }
    
    //assert(satsolver_sat(abstraction->sat) == SATSOLVER_SATISFIABLE);
}

ClausalAbstraction* clausal_abstractions_init(Solver* solver, Scope* scope, ClausalAbstraction* prev) {
    ClausalAbstraction* abstraction = malloc(sizeof(ClausalAbstraction));
    abstraction->sat = satsolver_init();
    abstraction->negation = satsolver_init();
    abstraction->result = QBF_RESULT_UNKNOWN;
    
    abstraction->b_lits = int_vector_init();
    abstraction->t_lits = int_vector_init();
    abstraction->entry = int_vector_init();
    abstraction->min_entry = int_vector_init();
    abstraction->assumptions = int_vector_init();
    abstraction->local_refinement_literals = NULL;
    if (scope->qtype == QUANTIFIER_EXISTENTIAL) {
        abstraction->local_refinement_literals = int_vector_init();
    }
    
    // strong unsat refinement
    abstraction->t_lit_mapping = map_init();
    abstraction->aux_var = 2 * solver->matrix->num_clauses + solver->matrix->max_vars + 1;
    assert(solver->matrix->last_clause == NULL || create_lit(abstraction->aux_var, false) > clause_to_b_lit(solver, solver->matrix->last_clause));
    abstraction->refinement = int_vector_init();
    abstraction->conjunctive_lits = int_vector_init();
    
    abstraction->scope = scope;
    abstraction->solver = solver;
    abstraction->prev = prev;
    abstraction->next = NULL;
    if (scope->num_next > 0) {
        abstraction->next = calloc(scope->num_next, sizeof(ClausalAbstraction*));
    }
    
    abstraction->statistics = statistics_init(10000);
    
#ifdef PARALLEL_SOLVING
    pthread_mutex_init(&abstraction->mutex, NULL);
    semaphore_init(&abstraction->sub_finished, 0);
    abstraction->num_started = 0;
#endif
    
    /*if (abstraction->next != NULL) {
        
    }*/
    //satsolver_save_original_clauses(abstraction->sat);
    
    if (scope->qtype == QUANTIFIER_EXISTENTIAL) {
        build_existential_abstraction(solver, abstraction, true);
        build_dual_universal_abstraction(solver, abstraction);
    } else {
        build_universal_abstraction(solver, abstraction);
        build_existential_abstraction(solver, abstraction, false);
    }
    
    assert(int_vector_is_sorted(abstraction->t_lits));
    assert(int_vector_is_sorted(abstraction->b_lits));
    
    assert(solver->matrix->last_clause == NULL || clause_to_t_lit(solver, solver->matrix->last_clause) < clause_to_b_lit(solver, solver->matrix->first_clause));
    assert(solver->matrix->last_clause == NULL || create_lit(solver->matrix->max_vars, false) < clause_to_t_lit(solver, solver->matrix->first_clause));
    
    if (prev != NULL) {
        for (size_t i = 0 ; i < int_vector_count(abstraction->t_lits); i++) {
            lit_t t_lit = int_vector_get(abstraction->t_lits, i);
            if (prev->scope->qtype == QUANTIFIER_EXISTENTIAL) {
                lit_t b_lit = t_lit_to_b_lit(solver, t_lit);
                assert(int_vector_contains_sorted(prev->b_lits, b_lit));
            } else {
                assert(int_vector_contains_sorted(prev->b_lits, t_lit));
            }
        }
    }
    
    return abstraction;
}


/*

Clausal_abstraction* create_clausal_abstraction(Caqe* solver, Scope* scope, vector* vars, vector* clauses, int level) {
    Clausal_abstraction* ca = malloc(sizeof(Clausal_abstraction));
    ca->scope = scope;
    ca->sat = satsolver_init();
    satsolver_save_original_clauses(ca->sat);
    ca->t_lits = int_vector_init();
    ca->b_lits = int_vector_init();
    if (scope->qtype == CAQE_QUANTIFIER_EXISTENTIAL) {
        ca->universal_quant = vector_init();
        ca->t_lit_quant = map_init();
        ca->instance_t_lits = NULL;
    } else {
        ca->existential_quant = NULL;
        ca->instance_t_lits = int_vector_init();
        ca->t_lit_quant = NULL;
    }
    ca->vars = vars;
    ca->clauses = clauses;
    ca->entry = int_vector_init();
    ca->greedyset = int_vector_init();
    ca->level = level;
    if (solver->options->strong_unsat_refinement) {
        ca->t_lit_mapping = map_init();
    } else {
        ca->t_lit_mapping = NULL;
    }
    ca->aux_vars = 2 * (solver->matrix->clause_num) + solver->matrix->var_num + 1;
    
    // Adjust size of SAT solver
    //satsolver_adjust(ca->sat, ca->aux_vars);
    
    
    // Build SAT instance
    for (size_t i = 0; i < vector_count(clauses); i++) {
        Clause* clause = vector_get(clauses, i);
        int t_lit = clause_to_t_lit(solver, clause);
        if (scope->qtype == CAQE_QUANTIFIER_UNIVERSAL && clause->min_scope->scope_id <= scope->scope_id && clause->max_scope->scope_id >= scope->scope_id) {
            int_vector_add(ca->instance_t_lits, t_lit);
        }
        for (Occ* occ = clause->first_occ; occ != NULL; occ = occ->next_in_clause) {
            if (occ->var->scope != scope) {
                continue;
            }
            if (occ->var->scope->qtype == CAQE_QUANTIFIER_UNIVERSAL) {
                satsolver_add(ca->sat, -t_lit);
                satsolver_add(ca->sat, -matrix_get_lit(occ));
                satsolver_add(ca->sat, 0);
                if (int_vector_find_sorted(ca->t_lits, t_lit) == VECTOR_NOT_FOUND) {
                    int_vector_add(ca->t_lits, t_lit);
                    
                    if (solver->options->default_polarities == CAQE_SET_POLARITY_FAST_SAT) {
                        satsolver_set_default_phase_lit(ca->sat, t_lit, -1);
                    } else if (solver->options->default_polarities == CAQE_SET_POLARITY_ALL ||
                               (solver->options->default_polarities == CAQE_SET_POLARITY_2QBF && ca->scope->next->next == NULL)) {
                        // TODO: scope->next->next ignores miniscoping
                        // last universal level
                        satsolver_set_default_phase_lit(ca->sat, t_lit, 1);
                    }
                }
            } else {
                assert(occ->var->scope->qtype == CAQE_QUANTIFIER_EXISTENTIAL);
                
                satsolver_add(ca->sat, matrix_get_lit(occ));
                if (scope->scope_id == 0 && solver->options->random_initial_assignment) {
                    if (occ->var->num_occ_neg > occ->var->num_occ_pos) {
                        satsolver_set_default_phase_lit(ca->sat, occ->var->var_id, -1);
                    } else if (occ->var->num_occ_neg < occ->var->num_occ_pos) {
                        satsolver_set_default_phase_lit(ca->sat, occ->var->var_id, 1);
                    } else {
                        if (rand() % 2 == 0) {
                            satsolver_set_default_phase_lit(ca->sat, occ->var->var_id, -1);
                        } else {
                            satsolver_set_default_phase_lit(ca->sat, occ->var->var_id, 1);
                        }
                    }
                }
                if (int_vector_find_sorted(ca->t_lits, t_lit) == VECTOR_NOT_FOUND) {
                    satsolver_add(ca->sat, -t_lit);
                    int_vector_add(ca->t_lits, t_lit);
                    if (solver->options->default_polarities != CAQE_SET_POLARITY_FAST_SAT &&
                        (solver->options->default_polarities == CAQE_SET_POLARITY_ALL ||
                         (solver->options->default_polarities == CAQE_SET_POLARITY_2QBF && ca->scope->next == NULL))) {
                            // last existential level
                            satsolver_set_default_phase_lit(ca->sat, t_lit, 1);
                        }
                }
            }
        }
        
        // Add t- and b-lits to existential quantifiers:
        // * We add t-lit if scope is between min- and max-scope of current clause
        // * We add b-lit if scope is between min- and max-scope of current clause, excluding max-scope
        if (scope->qtype == CAQE_QUANTIFIER_EXISTENTIAL && scope->scope_id >= clause->min_scope->scope_id && scope->scope_id <= clause->max_scope->scope_id) {
            int b_lit = t_lit_to_b_lit(solver, t_lit);
            if (scope == clause->max_scope) {
                if (int_vector_find_sorted(ca->t_lits, t_lit) == VECTOR_NOT_FOUND) {
                    // t-lit must be included in the lowest existential quantifier
                    satsolver_add(ca->sat, -t_lit);
                    int_vector_add(ca->t_lits, t_lit);
                    if (solver->options->default_polarities != CAQE_SET_POLARITY_FAST_SAT &&
                        (solver->options->default_polarities == CAQE_SET_POLARITY_ALL ||
                         (solver->options->default_polarities == CAQE_SET_POLARITY_2QBF && ca->scope->next == NULL))) {
                            // last existential level
                            satsolver_set_default_phase_lit(ca->sat, t_lit, 1);
                        }
                }
                satsolver_add(ca->sat, 0);
            } else if (int_vector_find_sorted(ca->t_lits, t_lit) != VECTOR_NOT_FOUND) {
                // have a t-lit in clause
                if (scope->scope_id < clause->max_scope->scope_id) {
                    satsolver_add(ca->sat, b_lit);
                    int_vector_add(ca->b_lits, b_lit);
                    if (solver->options->default_polarities != CAQE_SET_POLARITY_FAST_SAT && solver->options->default_polarities == CAQE_SET_POLARITY_ALL) {
                        satsolver_set_default_phase_lit(ca->sat, b_lit, -1);
                    }
                }
                satsolver_add(ca->sat, 0);
            }
        }
    }
    if (logging_get_verbosity() >= VERBOSITY_ALL) {
        logging_debug("level %d\n", ca->level);
        satsolver_print(ca->sat);
        
        int_vector_print(ca->t_lits);
    }
    
    assert(satsolver_sat(ca->sat) == SATSOLVER_SATISFIABLE);
    
    return ca;
}

void init_clausal_abstraction_for_scope(Caqe* solver, Scope* scope, vector* current_clauses, Clausal_abstraction* prev, int_vector* current_vars, int level) {
    if (scope == NULL) {
        return;
    }
    if (scope->qtype == CAQE_QUANTIFIER_UNIVERSAL) {
        // We split instances on the universal level
        assert(prev != NULL);
        assert(scope->next);
        size_t offset = 0;
        for (size_t i = 0; i < vector_count(scope->next->groups); i++) {
            int_vector* group = vector_get(scope->next->groups, i);
            int var = int_vector_get(group, 0);
            if (int_vector_contains_sorted(current_vars, var)) {
                // Group is relevant
                vector* clauses = vector_init();
                for (size_t j = 0; j < vector_count(current_clauses); j++) {
                    Clause* clause = vector_get(current_clauses, j);
                    bool contained = false;
                    for (Occ* occ = clause->first_occ; occ != NULL; occ = occ->next_in_clause) {
                        if (int_vector_contains_sorted(group, occ->var->var_id)) {
                            contained = true;
                            break;
                        }
                    }
                    if (contained) {
                        vector_add(clauses, clause);
                        map_add(prev->t_lit_quant, clause_to_t_lit(solver, clause), (void*)(intptr_t)(i-offset));
                    }
                }
                
                Clausal_abstraction* universal_quant = create_clausal_abstraction(solver, scope, scope->vars, clauses, level);
                vector_add(prev->universal_quant, universal_quant);
                init_clausal_abstraction_for_scope(solver, scope->next, universal_quant->clauses, universal_quant, group, level+1);
            } else {
                offset++;
            }
        }
    } else { // scope->qtype == CAQE_QUANTIFIER_EXISTENTIAL
        if (prev == NULL) {
            for (size_t i = 0; i < vector_count(scope->groups); i++) {
                // Top-level split, there is no universal level above
                int_vector* group = vector_get(scope->groups, i);
                
                vector* vars = vector_init();
                for (size_t j = 0; j < int_vector_count(group); j++) {
                    int var_id = int_vector_get(group, j);
                    Var* var = matrix_get_var(solver->matrix, var_id);
                    if (var->scope == scope) {
                        vector_add(vars, var);
                    }
                }
                
                vector* clauses = vector_init();
                for (size_t j = 0; j < vector_count(current_clauses); j++) {
                    Clause* clause = vector_get(current_clauses, j);
                    bool contained = false;
                    for (Occ* occ = clause->first_occ; occ != NULL; occ = occ->next_in_clause) {
                        if (int_vector_contains_sorted(group, occ->var->var_id)) {
                            contained = true;
                            break;
                        }
                    }
                    if (contained) {
                        vector_add(clauses, clause);
                    }
                }
                Clausal_abstraction* existential_quant = create_clausal_abstraction(solver, scope, vars, clauses, level);
                solver->entries[i] = existential_quant;
                init_clausal_abstraction_for_scope(solver, scope->next, existential_quant->clauses, existential_quant, group, level+1);
            }
        } else {
            vector* vars = vector_init();
            for (size_t j = 0; j < int_vector_count(current_vars); j++) {
                int var_id = int_vector_get(current_vars, j);
                Var* var = matrix_get_var(solver->matrix, var_id);
                if (var->scope == scope) {
                    vector_add(vars, var);
                }
            }
            
            Clausal_abstraction* existential_quant = create_clausal_abstraction(solver, scope, vars, current_clauses, level);
            prev->existential_quant = existential_quant;
            init_clausal_abstraction_for_scope(solver, scope->next, current_clauses, existential_quant, current_vars, level+1);
        }
    }
}

*/


/**
 * Returns true, if the lits contained in rhs is a subset of the lits in lhs
 * (regarding the variables lower than current scope).
 * Note that literals in clauses are sorted by their value.
 */
/*bool clause_is_subset(Clause* lhs, Clause* rhs, Scope* current_scope) {
    if (lhs == rhs) {
        return true;
    }
    Occ* rhs_occ = rhs->first_occ;
    Occ* lhs_occ = lhs->first_occ;
    while (lhs_occ != NULL) {
        if (rhs_occ == NULL) {
            return true;
        }
        Scope* rhs_scope = rhs_occ->var->scope;
        
        if (matrix_get_lit(lhs_occ) == matrix_get_lit(rhs_occ)) {
            // same literals, proceed with both
            lhs_occ = lhs_occ->next_in_clause;
            rhs_occ = rhs_occ->next_in_clause;
        } else if (matrix_get_lit(lhs_occ) < matrix_get_lit(rhs_occ)) {
            // proceed with next lit in lhs
            lhs_occ = lhs_occ->next_in_clause;
        } else {
            // lhs_occ->lit > rhs_occ->lit
            // proceed with rhs if variable is not existential and lower level
            if (rhs_scope->scope_id > current_scope->scope_id) {
                return false;
            }
            rhs_occ = rhs_occ->next_in_clause;
        }
    }
    
    while (rhs_occ != NULL) {
        // End of left hand side clause
        Scope* rhs_scope = rhs_occ->var->scope;
        if (rhs_scope->scope_id > current_scope->scope_id) {
            return false;
        }
        rhs_occ = rhs_occ->next_in_clause;
    }
    return true;
}*/

/**
 * Excludes the reason for the current unsat core from scope.
 * Returns false if no SAT call is needed, i.e., we can directly
 * push the unsat core to the scope above.
 */
/*bool unsat_refinement(Caqe* solver, Clausal_abstraction* quant, Clausal_abstraction* lower) {
    // check if an existential variable from this level is contained in refinement
    // if not, we don't need to do a SAT call and immediately return UNSAT
    
    assert(quant->scope->qtype == CAQE_QUANTIFIER_EXISTENTIAL);
    
    Scope* scope = quant->scope;
    
    bool need_sat_call = false;
    for (size_t j = 0; j < int_vector_count(solver->unsat_core); j++) {
        int t_lit = int_vector_get(solver->unsat_core, j);
        Clause* clause = matrix_get_clause(solver->matrix, t_lit_to_clause_id(solver, t_lit));
        for (Occ* occ = clause->first_occ; occ != NULL; occ = occ->next_in_clause) {
            Scope* var_scope = occ->var->scope;
            if (var_scope == scope) {
                need_sat_call = true;
                break;
            }
        }
    }
    
    if (!need_sat_call && !solver->options->certify) {
        // We do not need to refine because the entry will be excluded by a higher scope
        return false;
    }
    
    // One of the b-lits that were positive in the last iteration
    // must be negative in the next iteration
    int_vector* to_add = int_vector_init();
    logging_info("refinement: ");
    if (!solver->options->strong_unsat_refinement) {
        // Refine with UNSAT core of lower existential level
        int_vector* failed_assumptions = solver->unsat_core;
        assert(failed_assumptions != NULL);
        for (size_t i = 0; i < int_vector_count(failed_assumptions); i++) {
            int t_lit = int_vector_get(failed_assumptions, i);
            int b_lit = t_lit_to_b_lit(solver, t_lit);
            // do not refine if t-lit not contained on this level (and all levels above)
            Clause* clause = matrix_get_clause(solver->matrix, t_lit_to_clause_id(solver, t_lit));
            if (clause->min_scope->scope_id > scope->scope_id) {
                // t-lit not contained on this level (and all levels above)
                continue;
            }
            if (int_vector_find_sorted(quant->b_lits, b_lit) == VECTOR_NOT_FOUND) {
                int_vector_add(to_add, t_lit);
            }
            satsolver_add(quant->sat, -b_lit);
            logging_put(VERBOSITY_HIGH, "%d ", -b_lit);
        }
        
        satsolver_add(quant->sat, 0);
        logging_put(VERBOSITY_HIGH, "\n");
        
    } else if (solver->options->expansion_refinement && !solver->options->certify
               && solver->matrix->previous_scope == quant->scope->next->next
               && quant->level == 0) {
        // Expansion based refinment restricted to the lowest level
        // Applies only for 3QBF
        // 1. find the entry that was unsatisfiable
        Clausal_abstraction* inner_existential = lower->existential_quant;
        int_vector* entry = inner_existential->entry;
        Scope* inner_scope = inner_existential->scope;
        
        // 2. refine with the clausal abstraction of the last scope restricted to the complement of the entry.
        
        // map storing the relationship between auxilliary variables and those that they encode
        map* vars_to_refinement_vars = map_init();
        Matrix* refinement = matrix_init();
        int_vector* b_lits = int_vector_init();
        
        for (size_t i = 0; i < int_vector_count(inner_existential->t_lits); i++) {
            int t_lit = int_vector_get(inner_existential->t_lits, i);
            Clause* clause = matrix_get_clause(solver->matrix, t_lit_to_clause_id(solver, t_lit));
            
            if (int_vector_contains_sorted(entry, t_lit)) {
                continue;
            }
            
            if (clause->max_scope->scope_id < inner_scope->scope_id) {
                continue;
            }
            
            int b_lit = t_lit_to_b_lit(solver, t_lit);
            if (int_vector_find_sorted(quant->b_lits, b_lit) == VECTOR_NOT_FOUND) {
                int_vector_add(to_add, t_lit);
            }
            Clause* c_refine = matrix_new_clause(refinement);
            matrix_add_lit_to_clause(refinement, c_refine, - b_lit);
            
            // disable pure literal detection to prevent "overoptimization"
            int_vector_add(b_lits, b_lit);
            
            for (Occ* occ = clause->first_occ; occ != NULL; occ = occ->next_in_clause) {
                if (occ->var->scope == inner_scope) {
                    int var = occ->var->var_id;
                    if (!map_contains(vars_to_refinement_vars, var)) {
                        map_add(vars_to_refinement_vars, var, (void*)(intptr_t)quant->aux_vars++);
                    }
                    int lit = (int)(intptr_t)map_get(vars_to_refinement_vars, var);
                    assert(lit > 0);
                    if (lit > refinement->var_num) {
                        matrix_adjust(refinement, 2 * lit);
                    }
                    lit = occ->neg ? - lit : lit;
                    matrix_add_lit_to_clause(refinement, c_refine, lit);
                }
            }
            matrix_close_clause(refinement, c_refine);
        }
        
        // The upper loop would misses clauses that have no t_lits
        for (size_t i = 0; i < vector_count(inner_existential->clauses); i++) {
            Clause* clause = vector_get(inner_existential->clauses, i);
            if (clause->max_scope == clause->min_scope) {
                Clause* c_refine = matrix_new_clause(refinement);
                for (Occ* occ = clause->first_occ; occ != NULL; occ = occ->next_in_clause) {
                    if (occ->var->scope == inner_scope) {
                        int var = occ->var->var_id;
                        if (!map_contains(vars_to_refinement_vars, var)) {
                            map_add(vars_to_refinement_vars, var, (void*)(intptr_t)quant->aux_vars++);
                        }
                        int lit = (int)(intptr_t)map_get(vars_to_refinement_vars, var);
                        lit = occ->neg ? - lit : lit;
                        matrix_add_lit_to_clause(refinement, c_refine, lit);
                    }
                }
                matrix_close_clause(refinement, c_refine);
            }
        }
        map_free(vars_to_refinement_vars);
        
        
        //disable_pure_literal_detection
        for (size_t i = 0; i < int_vector_count(b_lits); i++) {
            int b_lit = int_vector_get(b_lits, i);
            Var* b_lit_var = matrix_get_var(refinement, b_lit);
            b_lit_var->disable_pure_literal_detection = true;
        }
        
        matrix_simplify(refinement);
        logging_info("Propagated %zu variables before refinement.\n", solver->matrix->propagation_queue_pos);
        
        for (size_t i = 0; i < int_vector_count(b_lits); i++) {
            int b_lit = int_vector_get(b_lits, i);
            Var* b_lit_var = matrix_get_var(refinement, b_lit);
            assert(b_lit_var->value != 1);
            if (b_lit_var->value == -1) {
                satsolver_add(quant->sat, - b_lit);
                satsolver_add(quant->sat, 0);
            }
        }
        int_vector_free(b_lits);
        
        if (!refinement->conflict) { // MATRIX_UNSATISFIABLE, MATRIX_UNKNOWN, or MATRIX_SATISFIABLE
            for (Clause* c = refinement->first_clause; c != NULL; c = c->next) {
                for (Occ* o = c->first_occ; o != NULL; o = o->next_in_clause) {
                    int lit = o->neg ? - o->var->var_id : o->var->var_id;
                    satsolver_add(quant->sat, lit);
                }
                satsolver_add(quant->sat, 0);
            }
        } else if (refinement->conflict) {
            // make the refinement false
            satsolver_add(quant->sat, 0);
        } else { // refinement->result == MATRIX_SATISFIABLE
            assert(false);
        }
        
        matrix_free(refinement);
        
    } else {

        // Strong unsat refinement: Exclude equivalent UNSAT cores
        assert(solver->options->strong_unsat_refinement);
        
        int_vector* refinement = int_vector_init();
        int_vector* conjunctive_lits = int_vector_init();
        int_vector* failed_assumptions = solver->unsat_core;
        assert(failed_assumptions != NULL);
        for (size_t i = 0; i < int_vector_count(failed_assumptions); i++) {
            int t_lit = int_vector_get(failed_assumptions, i);
            assert(!int_vector_contains_sorted(lower->existential_quant->entry, t_lit));
            // do not refine if t-lit not contained on this level (and all levels above)
            Clause* clause = matrix_get_clause(solver->matrix, t_lit_to_clause_id(solver, t_lit));
            if (clause->min_scope->scope_id > scope->scope_id) {
                // t-lit not contained on this level (and all levels above)
                continue;
            }
            
            // Was the t-lit processed before?
            // TODO: disabling caching is not optimal (proof could save the auxilliary variables as well)
            if (!solver->options->certify && map_contains(quant->t_lit_mapping, t_lit)) {
                int lit = (int)(intptr_t)map_get(quant->t_lit_mapping, t_lit);
                int_vector_add(refinement, lit);
                continue;
            }
            
            // Get some random existential occurrence from clause, so we can use
            // the occurrence list to iterate over clauses
            int_vector_reset(conjunctive_lits);
            for (Occ* occ = clause->first_occ; occ != NULL; occ = occ->next_in_clause) {
                Scope* var_scope = occ->var->scope;
                
                // Consider only existential variables that have a lower level
                if (var_scope->qtype == CAQE_QUANTIFIER_UNIVERSAL || var_scope->scope_id <= scope->scope_id) {
                    continue;
                }
                
                // Iterate over occurrence list and add equivalent clauses
                Occ* var_occ = !occ->neg ? occ->var->first_pos : occ->var->first_neg;
                for (; var_occ != NULL; var_occ = var_occ->next_in_occ) {
                    int b_lit = clause_to_b_lit(solver, var_occ->clause);
                    if (int_vector_contains_sorted(conjunctive_lits, b_lit)) {
                        continue;
                    }
                    // clause is subset is probably more expensive that sorted find, that's why it is guarded here
                    if (clause_is_subset(clause, var_occ->clause, scope)) {
                        int_vector_add_sorted(conjunctive_lits, b_lit);
                    }
                }
            }
            
            assert(int_vector_count(conjunctive_lits) > 0);
            
            //int_vector_print(conjunctive_lits);
            assert(int_vector_count(conjunctive_lits) > 0);
            if (int_vector_count(conjunctive_lits) == 1) {
                // do not need auxilliary variable
                int b_lit = int_vector_get(conjunctive_lits, 0);
                if (int_vector_find_sorted(quant->b_lits, b_lit) == VECTOR_NOT_FOUND) {
                    int_vector_add(to_add, b_lit_to_t_lit(solver, b_lit));
                }
                int_vector_add(refinement, -b_lit);
                if (!solver->options->certify) {
                    map_add(quant->t_lit_mapping, t_lit, (void*)(intptr_t)-b_lit);
                }
            } else {
                // Build the conjunction using an auxilliary variable
                int aux_var = quant->aux_vars++;
                int_vector_add(refinement, aux_var);
                if (!solver->options->certify) {
                    map_add(quant->t_lit_mapping, t_lit, (void*)(intptr_t)aux_var);
                }
                
                for (size_t i = 0; i < int_vector_count(conjunctive_lits); i++) {
                    int b_lit = int_vector_get(conjunctive_lits, i);
                    if (int_vector_find_sorted(quant->b_lits, b_lit) == VECTOR_NOT_FOUND) {
                        int_vector_add(to_add, b_lit_to_t_lit(solver, b_lit));
                    }
                    
                    satsolver_add(quant->sat, -aux_var);
                    satsolver_add(quant->sat, -b_lit);
                    satsolver_add(quant->sat, 0);
                }
                
                certification_strengthen_conjunction(solver, conjunctive_lits);
            }
        }
        
        for (size_t i = 0; i < int_vector_count(refinement); i++) {
            int lit = int_vector_get(refinement, i);
            satsolver_add(quant->sat, lit);
            logging_put(VERBOSITY_HIGH, "%d ", lit);
        }
        satsolver_add(quant->sat, 0);
        logging_put(VERBOSITY_HIGH, "\n");
        
        int_vector_free(refinement);
        int_vector_free(conjunctive_lits);
    }
    
    for (size_t i = 0; i < int_vector_count(to_add); i++) {
        int t_lit = int_vector_get(to_add, i);
        int b_lit = t_lit + solver->matrix->clause_num;
        satsolver_add(quant->sat, -t_lit);
        satsolver_add(quant->sat, b_lit);
        satsolver_add(quant->sat, 0);
        int_vector_add_sorted(quant->t_lits, t_lit);
        int_vector_add_sorted(quant->b_lits, b_lit);
    }
    int_vector_free(to_add);
    
    return need_sat_call;
}*/

/**
 * Excludes the reason for the current unsat core from scope.
 * Returns false if no SAT call is needed, i.e., we can directly
 * push the unsat core to the scope above.
 */
/*bool sat_refinement(Caqe* solver, Clausal_abstraction* quant) {
    // check if an universal variable from this level is contained in refinement
    // if not, we don't need to do a SAT call and immediately return SAT
    
    assert(quant->scope->qtype == CAQE_QUANTIFIER_UNIVERSAL);
    
    Scope* scope = quant->scope;
    int_vector* sub_entry = quant->existential_quant->entry;
    
    bool need_sat_call = false;
    for (size_t j = 0; j < int_vector_count(sub_entry); j++) {
        int t_lit = int_vector_get(sub_entry, j);
        Clause* clause = matrix_get_clause(solver->matrix, t_lit_to_clause_id(solver, t_lit));
        for (Occ* occ = clause->first_occ; occ != NULL; occ = occ->next_in_clause) {
            Scope* var_scope = occ->var->scope;
            if (var_scope == scope) {
                need_sat_call = true;
                break;
            }
        }
    }
    
    if (!need_sat_call) {
        // We do not need to refine because the entry will be excluded by a higher scope
        int_vector_reset(quant->entry);
        for (size_t j = 0; j < int_vector_count(sub_entry); j++) {
            int t_lit = int_vector_get(sub_entry, j);
            int_vector_add(quant->entry, t_lit);
        }
    } else {
        logging_info("refinement ");
        // Some clause in the entry must be satisfied by player X
        for (size_t i = 0; i < int_vector_count(sub_entry); i++) {
            int t_lit = int_vector_get(sub_entry, i);
            int_vector_add_sorted(quant->t_lits, t_lit);  // only adds it if it is not already contained
            assert(int_vector_contains_sorted(quant->instance_t_lits, t_lit));
            satsolver_add(quant->sat, t_lit);
            logging_put(VERBOSITY_ALL, "%d ", t_lit);
        }
        satsolver_add(quant->sat, 0);
        logging_put(VERBOSITY_ALL, "\n");
    }
    return need_sat_call;
}*/

/**
 * @abstract Removes t-lits from entry according to current assignment (assumes last call was SAT).
 * @return true if entry was changed.
 */
/*bool entry_minimization(Caqe* solver, Clausal_abstraction* quant, int_vector* entry, int_vector* deleted, bool greedy) {
    logging_debug("Greedy entry minimization.\n");
    SATSolver* sat = quant->sat;
    int elim = 0;
    int greedy_num = 0;
    
    solver->num_minimizations++;
    
    for (size_t i = 0; i < int_vector_count(entry); i++) {
        int t_lit = int_vector_get(entry, i);
        Clause* clause = matrix_get_clause(solver->matrix, t_lit_to_clause_id(solver, t_lit));
        clause->is_in_entry = solver->num_minimizations;
    }
    
    for (size_t i = 0; i < int_vector_count(entry); i++) {
        int t_lit = int_vector_get(entry, i);
        Clause* clause = matrix_get_clause(solver->matrix, t_lit_to_clause_id(solver, t_lit));
        logging_debug("Checking for greedy assignments in clause %d.\n", clause->clause_id);
        if (clause->is_in_entry != solver->num_minimizations) {
            assert(clause->is_in_entry == -1);
            // was eliminated by a greedy assignment
            continue;
        }
        
        Var* max_var = NULL;
        int max_lit = 0;
        int max_occs = 0;
        
        for (Occ* o = clause->first_occ; o != NULL; o = o->next_in_clause) {
            int lit = matrix_get_lit(o);
            Scope* var_scope = o->var->scope;
            // Check whether lit can be greedily assigned in clause
            if (var_scope != quant->scope) {
                // we can only assign variables in the same quantifier
                continue;
            }
            
            int assignment = satsolver_deref_partial(sat, lit);
            if (assignment > 0) {
                // OK, clause is satisfied, we can drop this clause (t-lit) from entry
                clause->is_in_entry = -1;
                elim++;
                break;
            } else if (assignment == 0) {
                // not yet assigned, assign it greedily to satisfy clause
                if (int_vector_find_sorted(quant->greedyset, lit) != VECTOR_NOT_FOUND) {
                    // lit is already set greedily
                    clause->is_in_entry = -1;
                    elim++;
                    break;
                } else if (int_vector_find_sorted(quant->greedyset, -lit) == VECTOR_NOT_FOUND) {
                    // negation is not yet set greedily
                    // we can set the literal! when should we? count how many clauses it would eliminate
                    Occ* occs = lit > 0 ? o->var->first_pos : o->var->first_neg;
                    int occs_in_entry = 0;
                    for (; occs != NULL; occs = occs->next_in_occ) {
                        if (occs->clause->is_in_entry == solver->num_minimizations) {
                            occs_in_entry++;
                        }
                    }
                    logging_debug("Lit %d would eliminate %d clauses.\n", lit, occs_in_entry);
                    if (greedy && occs_in_entry > max_occs) {
                        logging_debug("Found a better literal to eliminate: %d better than %d; will eliminate %d clauses from the entry.\n", lit, max_lit, max_occs);
                        max_var = o->var;
                        max_occs = occs_in_entry;
                        max_lit = lit;
                    }
                }
            }
        }
        
        if (max_var != NULL) {
            logging_debug("Greedily setting %d (%d); eliminates %d clauses from the entry (round %d).\n", max_lit, max_var->var_id, max_occs, solver->num_minimizations);
            assert(lit_to_var(max_lit) == max_var->var_id);
            greedy_num++;
            int_vector_add_sorted(quant->greedyset, max_lit);
            
            // Mark all clauses in the occurrence list as not part of the entry any more.
            Occ* occs = max_lit > 0 ? max_var->first_pos : max_var->first_neg;
            for (; occs != NULL; occs = occs->next_in_occ) {
                if (occs->clause->is_in_entry == solver->num_minimizations) {
                    occs->clause->is_in_entry = -1;
                    elim++;
                }
            }
        }
    }
    
    logging_info("Eliminated %d literals via cheap strategy.\n", elim);
    logging_debug("Greedy Y-assignments: %d\n", greedy_num);
    
    assert(greedy || greedy_num == 0);
    
    // cleanup deleted elements
    for (size_t i = 0; i < int_vector_count(entry); i++) { // FIME: performance bug. Quadratic time in the size of the entry.
        int t_lit = int_vector_get(entry, i);
        Clause* clause = matrix_get_clause(solver->matrix, t_lit_to_clause_id(solver, t_lit));
        if (clause->is_in_entry == -1) {
            // deleted
            if (deleted != NULL) {
                int_vector_add_sorted(deleted, t_lit); // FIME: performance bug
            }
            int_vector_remove_index(entry, i); // FIME: performance bug.
            i--;
        }
    }
    return elim > 0;
}*/
