//
//  main.c
//  caqe
//
//  Copyright (c) 2015 - 2016, Leander Tentrup, Saarland University
//                2015, Markus Rabe, University of California, Berkeley
//
//  Licensed under Apache License v2.0, see LICENSE.txt for information
//

#include <stdio.h>
#include <stdbool.h>
#include <assert.h>
#include <string.h>

#include "certification.h"
#include "config.h"
#include "getopt.h"
#include "logging.h"
#include "satsolver.h"
#include "solver.h"
#include "qdimacs.h"

static void print_usage(const char* name) {
    printf("usage: %s [OPTIONS] [file]\n\n"
           "optional arguments:\n"
           "  -h, --help       show this help message and exit\n"
#ifdef CERTIFICATION
           "  -c               enable certification\n"
#endif
           "  --statistics     show detailed solving statistics\n"
#ifdef PARALLEL_SOLVING
           "  --num-threads N  number of threads to use during solving (default 2)\n"
#endif
           "  --version        show version and exit\n" , name);
}

static void print_version() {
    printf("CAQE %s\n\n", solver_get_version());
    printf("Copyright 2015 - 2016, Leander Tentrup, Saarland University\n"
           "Copyright 2015, Markus N. Rabe, University of California, Berkeley\n"
           "Licensed under Apache License v2.0\n");

    printf("\nThis software builds on top of the following projects:\n");
    printf("%s %s, %s\n", satsolver_get_name(), satsolver_get_version(), satsolver_copyright());
    printf("libcperciva getopt, Copyright 2005 - 2014, Colin Percival, 2014 Sean Kelly\n");
}

static bool parse_boolean_argument(const char* cmd, const char* arg) {
    if (strlen(arg) != 1 || (arg[0] != '1' && arg[0] != '0')) {
        logging_fatal("Wrong argument %s for %s, expect 0/1\n", arg, cmd);
    }
    return arg[0] == '1';
}

int main(int argc, char* const argv[]) {
    SolverOptions* options  = solver_get_default_options();
    
    const char* file_name = NULL;
    FILE* file = NULL;
    
    // Handling of command line arguments
    //opterr = 0; // Disable error messages as we handle them ourself
    const char * ch;
    while ((ch = GETOPT(argc, argv)) != NULL) {
        GETOPT_SWITCH(ch) {
            GETOPT_OPT("-h"):
            GETOPT_OPT("--help"):
            print_usage(argv[0]);
            return 0;
            GETOPT_OPT("-v"):
            logging_set_verbosity(VERBOSITY_ALL);
            break;
            GETOPT_OPT("--version"):
            print_version();
            return 0;
            
            GETOPT_OPT("-c"):
            options->certify = true;
            break;
            
            GETOPT_OPTARG("--preprocessing"):
            options->preprocess = parse_boolean_argument(ch, optarg);
            break;
            GETOPT_OPTARG("--miniscoping"):
            options->miniscoping = parse_boolean_argument(ch, optarg);
            break;
            GETOPT_OPT("--statistics"):
            options->statistics = true;
            break;
            
            GETOPT_OPTARG("--dual-propagation"):
            options->dual_propagation_minimization = parse_boolean_argument(ch, optarg);
            break;
            GETOPT_OPTARG("--strong-unsat-refinement"):
            options->strong_unsat_refinement = parse_boolean_argument(ch, optarg);
            break;
            
#ifdef PARALLEL_SOLVING
            GETOPT_OPTARG("--num-threads"):
            options->num_threads = strtoul(optarg, NULL, 0);
            if (options->num_threads == 0) {
                logging_error("Illegal number of threads argument %zu\n", options->num_threads);
                print_usage(argv[0]);
                return 1;
            }
            break;
#endif
        
        GETOPT_MISSING_ARG:
            printf("missing argument to %s\n", ch);
            /* FALLTHROUGH */
        GETOPT_DEFAULT:
            printf("unknown argument %s\n", ch);
            print_usage(argv[0]);
            return 1;
        }
    }
    
    if (options->certify) {
        fixme("Reenable optimizations for certification run\n");
        options->strong_unsat_refinement = false;
        options->miniscoping = false;
    }
    
    if (optind + 1 < argc) {
        logging_error("too many arguments, see --help for more information\n");
        return 1;
    }
    
    if (optind < argc) {
        file_name = argv[optind];
    } else {
        file = stdin;
    }
    
    Matrix* matrix = matrix_init();
    if (!matrix) {
        return 1;
    }
    
    Stats* parsing = statistics_init(1000);
    statistics_start_timer(parsing);
    
    int error;
    if (file == NULL) {
        error = qdimacs_open_and_read_qcir_file(matrix, file_name);
    } else {
        error = matrix_from_qdimacs(matrix, file);
    }
    if (error) {
        return 1;
    }
    
    statistics_stop_and_record_timer(parsing);
    
    Solver* solver = solver_init(matrix, options);
    if (!solver) {
        return 1;
    }
    
    qbf_res res = solver_sat(solver);
    
    if (!options->certify) {
        if (res == QBF_RESULT_SAT) {
            printf("SAT\n");
        } else if (res == QBF_RESULT_UNSAT) {
            printf("UNSAT\n");
        }
    } else {
#ifdef CERTIFICATION
        certification_print(&solver->certificate, res);
#endif
    }
    
    if (options->statistics) {
        printf("parsing took ");
        statistics_print_time(parsing);
        solver_print_statistics(solver);
    }
    statistics_free(parsing);
    
    assert(res != QBF_RESULT_UNKNOWN);
    
    return res;
}
