//
//  matrix.c
//  caqe
//
//  Copyright (c) 2015 - 2016, Leander Tentrup, Saarland University
//                2015, Markus Rabe, University of California, Berkeley
//
//  Licensed under Apache License v2.0, see LICENSE.txt for information
//

#include "matrix.h"

#include <assert.h>
#include <stdint.h>
#include <string.h>

#include "logging.h"
#include "matrix_check.h"
#include "matrix_print.h"

// Helper

static void set_size(void** buffer, size_t* size, size_t new_size, size_t bytes, size_t num) {
    assert(num <= *size);
    
    const size_t old_size = *size;
    assert(new_size > old_size);
    
    void* old_buffer = *buffer;
    void* new_buffer = calloc(new_size, bytes);
    if (num > 0) {
        // copy old values
        memcpy(new_buffer, old_buffer, num * bytes);
    }
    free(old_buffer);
    
    *size = new_size;
    *buffer = new_buffer;
}

static void enlarge(void** buffer, size_t* size, size_t bytes, size_t num) {
    assert(num <= *size);
    
    const size_t old_size = *size;
    const size_t new_size = old_size > 0 ? old_size * 2 : 1;
    
    set_size(buffer, size, new_size, bytes, num);
}

static void adjust_variables(Matrix* matrix, size_t max_var) {
    if (max_var < matrix->size_vars) {
        matrix->max_vars = max_var;
        return;
    }
    // use exponential growth strategy
    size_t new_size = (matrix->size_vars == 0) ? 1 : 2 * matrix->size_vars;
    while (new_size <= max_var) {
        new_size *= 2;
    }
    // TODO: check if casting is correct
    set_size((void**)&matrix->vars, &matrix->size_vars, new_size, sizeof(Var*), matrix->size_vars);
    matrix->max_vars = max_var;
}

static void adjust_clauses(Matrix* matrix, size_t max_clause) {
    if (max_clause < matrix->size_clauses) {
        return;
    }
    // use exponential growth strategy
    size_t new_size = (matrix->size_clauses == 0) ? 1 : 2 * matrix->size_clauses;
    while (new_size < max_clause) {
        new_size *= 2;
    }
    // TODO: check if casting is correct
    set_size((void**)&matrix->clauses, &matrix->size_clauses, new_size, sizeof(Clause*), matrix->num_clauses);
}

static void import_literal(Matrix* matrix, lit_t literal) {
    const var_t variable = lit_to_var(literal);
    if (variable >= matrix->size_vars) {
        adjust_variables(matrix, variable);
    }
    if (variable > matrix->max_vars) {
        matrix->max_vars = variable;
    }
}


// Constructors

Matrix* matrix_init() {
    Matrix* matrix = malloc(sizeof(Matrix));
    
    matrix->vars = NULL;
    matrix->max_vars = 0;
    matrix->size_vars = 0;
    
    matrix->first_clause = NULL;
    matrix->last_clause = NULL;
    
    matrix->clauses = NULL;
    matrix->num_clauses = 0;
    matrix->size_clauses = 0;
    
    matrix->last_scope_id = 0;
    matrix->previous_scope = NULL;
    matrix->top_level = matrix_new_scope(matrix, QUANTIFIER_EXISTENTIAL);
    
    matrix->conflict = false;
    int_queue_init(&matrix->propagation_queue);
    matrix->phase = PHASE_BUILDING;
    
    return matrix;
}

void matrix_adjust(Matrix* matrix, size_t num_vars, size_t num_clauses) {
    adjust_variables(matrix, num_vars);
    adjust_clauses(matrix, num_clauses);
}

void matrix_free(Matrix* matrix) {
    assert(matrix);
    fixme("matrix_free is not implemented\n");
}


// Variables

Var* matrix_new_var(Matrix* matrix, Scope* scope, var_t var_id) {
    api_expect(var_id > 0, "variables must be greater than 0");
    Var* var = malloc(sizeof(Var));
    
    var->var_id = var_id;
    var->orig_var_id = var_id;
    
    var->num_occ_pos = 0;
    var->first_pos = NULL;
    var->last_pos = NULL;
    
    var->num_occ_neg = 0;
    var->first_neg = NULL;
    var->last_neg = NULL;
    
    var->scope = scope;
    var->value = 0;
    var->removed = false;
    var->orig_quant = scope->qtype;
    int_queue_init(&var->reductions);
    
    vector_add(scope->vars, var);
    
    import_literal(matrix, var_id);
    
    matrix->vars[var_id] = var;
    
    return var;
}

Var* matrix_get_var(Matrix* matrix, var_t var_id) {
    api_expect(var_id > 0, "variables must be greater than 0\n");
    if (var_id > matrix->max_vars) {
        return NULL;
    }
    return matrix->vars[var_id];
}

// Scopes

static Scope* create_scope(Matrix* matrix, quant_t qtype) {
    Scope* scope = malloc(sizeof(Scope));
    scope->matrix = matrix;
    scope->scope_id = 0;
    scope->qtype = qtype;
    assert(qtype == QUANTIFIER_EXISTENTIAL || qtype == QUANTIFIER_UNIVERSAL);
    scope->vars = vector_init();
    
    scope->prev = NULL;
    scope->next = NULL;
    scope->num_next = 0;
    
    scope->first_clause = scope->last_clause = NULL;
    scope->group = 0;
    
    return scope;
}

static void link_scope(Scope* prev, Scope* next) {
    assert(prev != NULL);
    for (size_t i = 0; i < prev->num_next; i++) {
        if (prev->next[i] == next) {
            return;
        }
    }
    
    const size_t old_num_scopes = prev->num_next;
    prev->num_next++;
    Scope** new_next = malloc(prev->num_next * sizeof(Scope*));
    for (size_t i = 0; i < old_num_scopes; i++) {
        new_next[i] = prev->next[i];
    }
    new_next[prev->num_next - 1] = next;
    free(prev->next);
    prev->next = new_next;
    next->prev = prev;
}


static void unlink_scope(Scope* prev, Scope* current) {
    // Remove pointer to current from prev->scopes
    size_t j = 0;
    for (size_t i = 0; i < prev->num_next; i++) {
        if (prev->next[i] != current) {
            prev->next[j] = prev->next[i];
            j++;
        }
    }
    assert(j == prev->num_next - 1);
    prev->num_next = j;
    current->prev = NULL;
    
    if (prev->num_next == 0) {
        free(prev->next);
        prev->next = NULL;
    }
}

Scope* matrix_new_scope(Matrix* matrix, quant_t qtype) {
    if (matrix->previous_scope != NULL && matrix->previous_scope->qtype == qtype) {
        // Return previous scope if it has the same quantifier as requested
        return matrix->previous_scope;
    }
    
    Scope* scope = create_scope(matrix, qtype);
    scope->scope_id = matrix->last_scope_id++;
    assert((qtype == QUANTIFIER_EXISTENTIAL && scope->scope_id % 2 == 0)
           || (qtype == QUANTIFIER_UNIVERSAL && scope->scope_id % 2 == 1));
    
    if (matrix->previous_scope) {
        link_scope(matrix->previous_scope, scope);
    }
    matrix->previous_scope = scope;
    
    return scope;
}

Clause* matrix_iterator_start(const Matrix* matrix, const Scope* scope) {
    if (scope->first_clause != NULL) {
        return scope->first_clause;
    }
    return matrix->first_clause;
}

Clause* matrix_iterator_end(const Matrix* matrix, const Scope* scope) {
    assert(matrix);
    if (scope->last_clause) {
        return scope->last_clause->next;
    }
    return NULL;
}

// Clauses

static void append_clause(Matrix* matrix, Clause* clause) {
    clause->next = NULL;
    clause->prev = matrix->last_clause;
    
    if (matrix->last_clause != NULL) {
        matrix->last_clause->next = clause;
        matrix->last_clause = clause;
        assert(matrix->first_clause != NULL);
    } else {
        assert(matrix->first_clause == NULL);
        matrix->first_clause = clause;
        matrix->last_clause = clause;
    }
}

static void insert_after(Matrix* matrix, Clause* clause, Clause* insertion) {
    assert(clause != NULL);
    assert(clause->next == NULL);
    assert(insertion->prev == NULL);
    assert(clause != matrix->last_clause);
    
    clause->next = insertion;
    insertion->prev = clause;
}

Clause* matrix_new_clause(Matrix* matrix) {
    const var_t clause_id = matrix->num_clauses;
    if (clause_id >= matrix->size_clauses) {
        enlarge((void**)&matrix->clauses, &matrix->size_clauses, sizeof(Clause*), matrix->num_clauses);
    }
    matrix->num_clauses++;
    
    Clause* clause = malloc(sizeof(Clause));
    
    clause->clause_id = clause->orig_clause_id = clause_id;
    assert(matrix->clauses[clause->clause_id] == NULL);
    matrix->clauses[clause->clause_id] = clause;
    
    clause->size = 0;
    clause->first_occ = NULL;
    clause->last_occ = NULL;

    clause->is_in_entry = -1;
    clause->satisified = false;
    
    append_clause(matrix, clause);
    
    clause->scopes = bit_vector_init(0, matrix->last_scope_id);
    
    return clause;
}

static bool add_occurrence_to_clause(Clause* clause, Occ* occ) {
    
    // update linked list representing clause
    if (clause->last_occ == NULL) {
        // clause is empty
        assert(clause->size == 0);
        assert(clause->first_occ == NULL);
        
        occ->next_in_clause = NULL;
        occ->prev_in_clause = NULL;
        clause->last_occ = occ;
        clause->first_occ = occ;
    } else {
        assert(clause->first_occ != NULL);
        
        // literals in clauses are sorted, so use insertion sort
        const lit_t lit = matrix_get_lit(occ);
        const var_t var_id = lit_to_var(lit);
        Occ* other = NULL;
        for (other = clause->last_occ; other != NULL; other = other->prev_in_clause) {
            if (lit_to_var(matrix_get_lit(other)) < var_id) {
                break;
            } else if (lit_to_var(matrix_get_lit(other)) == var_id) {
                if (matrix_get_lit(other) == lit) {
                    // already contained in clause
                    free(occ);
                    return false;
                } else {
                    assert(matrix_get_lit(other) == -lit);
                    occ->clause->satisified = true;  // mark clause as trivially satisfied
                }
            }
        }
        if (other == NULL) {
            // insert at the beginning
            occ->prev_in_clause = NULL;
            occ->next_in_clause = clause->first_occ;
            clause->first_occ->prev_in_clause = occ;
            clause->first_occ = occ;
        } else {
            // insert after other
            if (other->next_in_clause == NULL) {
                // insert at the end
                occ->next_in_clause = NULL;
                occ->prev_in_clause = other;
                other->next_in_clause = occ;
                clause->last_occ = occ;
            } else {
                // in the middle
                occ->next_in_clause = other->next_in_clause;
                occ->prev_in_clause = other;
                other->next_in_clause = occ;
                occ->next_in_clause->prev_in_clause = occ;
            }
        }
    }
    
    clause->size += 1;
    
    // Update min/max scope
    bit_vector_add(clause->scopes, occ->var->scope->scope_id);
    assert(bit_vector_min(clause->scopes) != BIT_VECTOR_NO_ENTRY);
    
    return true;
}

static void add_to_occurrence_list(Var* var, Occ* occ) {
    const lit_t lit = matrix_get_lit(occ);
    if (lit > 0) {
        var->num_occ_pos += 1;
        if (var->first_pos == NULL) {
            assert(var->last_pos == NULL);
            occ->next_in_occ = NULL;
            occ->prev_in_occ = NULL;
            var->first_pos = occ;
            var->last_pos = occ;
        } else {
            assert(var->last_pos != NULL);
            occ->next_in_occ = NULL;
            occ->prev_in_occ = var->last_pos;
            var->last_pos->next_in_occ = occ;
            var->last_pos = occ;
        }
    } else {
        var->num_occ_neg += 1;
        if (var->first_neg == NULL) {
            assert(var->last_neg == NULL);
            occ->next_in_occ = NULL;
            occ->prev_in_occ = NULL;
            var->first_neg = occ;
            var->last_neg = occ;
        } else {
            assert(var->last_neg != NULL);
            occ->next_in_occ = NULL;
            occ->prev_in_occ = var->last_neg;
            var->last_neg->next_in_occ = occ;
            var->last_neg = occ;
        }
    }
}

Occ* matrix_add_lit_to_clause(Matrix* matrix, Clause* clause, lit_t lit) {
    Occ* occ = malloc(sizeof(Occ));
    import_literal(matrix, lit);
    const var_t var_id = lit_to_var(lit);
    Var* var = matrix->vars[var_id];
    if (var == NULL) { // unbound variable, put variable to outermost scope
        var = matrix_new_var(matrix, matrix->top_level, var_id);
    }
    //api_expect(var != NULL, "Variables must be bound\n");
    
    occ->neg = lit < 0;
    occ->var = var;
    occ->clause = clause;
    
    bool added = add_occurrence_to_clause(clause, occ);
    if (added) {
        add_to_occurrence_list(var, occ);
    } else {
        // could also return occ of contained lit
        return NULL;
    }
    
    return occ;
}

static void remove_occurrence_from_occurrence_list(Matrix*, Occ*);

static void unlink_clause_block(Matrix* matrix, Clause* first, Clause* last) {
#ifndef NDEBUG
    // check that first and last build a block
    for (Clause* clause = first; clause != last; clause = clause->next) {
        assert(clause != NULL);
    }
#endif
    
    if (matrix->first_clause == first) {
        matrix->first_clause = last->next;
    }
    if (matrix->last_clause == last) {
        matrix->last_clause = first->prev;
    }
    
    if (first->prev != NULL) {
        first->prev->next = last->next;
    }
    if (last->next != NULL) {
        last->next->prev = first->prev;
    }
    
    first->prev = NULL;
    last->next = NULL;
    
    assert(first->prev == NULL);
    assert(last->next == NULL);
}

/**
 * Unlinks clause from the linked list representing matrix
 */
static void unlink_clause(Matrix* matrix, Clause* clause) {
    unlink_clause_block(matrix, clause, clause);
}

static void remove_clause(Matrix* matrix, Clause* clause) {
    // The clause is satisfied and needs to be removed from all occurrence lists
    logging_debug("removing clause %d\n", clause->clause_id);
    Occ* next_in_clause = NULL;
    for (Occ* occs_in_clause = clause->first_occ; occs_in_clause != NULL; occs_in_clause = next_in_clause) {
        next_in_clause = occs_in_clause->next_in_clause;
        remove_occurrence_from_occurrence_list(matrix, occs_in_clause);
        free(occs_in_clause);
    }
    
    unlink_clause(matrix, clause);
    
    matrix->clauses[clause->clause_id] = NULL;
    
    free(clause);
}

static void check_propagation_from_clause(Matrix* matrix, Clause* clause) {
    // check properties of the closed clause:
    if (clause->size == 0) {
        logging_info("clause %d is empty\n", clause->clause_id);
        //assert(!matrix->conflict);
        matrix->conflict = true;
        bit_vector_reset(clause->scopes);
        return;
    }
    
    if (clause->size == 1) {
        // singleton clause
        lit_t lit = matrix_get_lit(clause->first_occ);
        var_t var_id = lit_to_var(lit);
        Var* var = matrix->vars[var_id];
        assert(var != NULL);
        
        if (var->scope->qtype == QUANTIFIER_UNIVERSAL) {
            logging_info("clause %d has a single universal variable\n", clause->clause_id);
            int value = lit < 0 ? 1 : -1;
            int_queue_push(&matrix->propagation_queue, create_lit_from_value(var->var_id, value));
            return;
        } else {
            int_queue_push(&matrix->propagation_queue, lit);
        }
    }
}

void matrix_close_clause(Matrix* matrix, Clause* clause) {
    
    check_propagation_from_clause(matrix, clause);
    
    if (clause->satisified) {
        logging_debug("clause %d is a tautology\n", clause->clause_id);
        remove_clause(matrix, clause);
    }
}


// Preprocessing

static void propagate_var_with_empty_occurrence_list(Matrix* matrix, Var* var) {
    if (matrix->phase != PHASE_PREPROCESSING) {
        return;
    }
    if (var->value != 0) {
        return;
    }
    int value = 0;
    if (var->num_occ_neg == 0) {
        // variable appears only positively
        value = 1;
    } else if (var->num_occ_pos == 0) {
        // variable appears only negatively
        value = -1;
    }
    if (var->scope->qtype == QUANTIFIER_UNIVERSAL) {
        value = -value;
    }
    if (value != 0) {
        logging_info("variable %d is pure, value %d\n", var->var_id, value);
        int_queue_push(&matrix->propagation_queue, create_lit_from_value(var->var_id, value));
    }
}

static void reduce_clause_universally(Matrix*, Clause*);

static void remove_occurrence_from_clause(Matrix* matrix, Occ* occ) {
    Clause* clause = occ->clause;
    logging_debug("removing %d from clause %d\n", matrix_get_lit(occ), clause->clause_id);
    
    assert(clause->size > 0);
    if (clause->first_occ == occ && clause->last_occ == occ) {
        // is the only literal
        clause->last_occ  = NULL;
        clause->first_occ = NULL;
        assert(clause->size == 1);
    } else if (clause->first_occ == occ) {
        // is the first occurrence, but not the last
        clause->first_occ = occ->next_in_clause;
        occ->next_in_clause->prev_in_clause = NULL;
    } else if(occ->clause->last_occ == occ) {
        // is the last occurrence, but not the first
        clause->last_occ = occ->prev_in_clause;
        occ->prev_in_clause->next_in_clause = NULL;
    } else { // is a middle node
        occ->prev_in_clause->next_in_clause = occ->next_in_clause;
        occ->next_in_clause->prev_in_clause = occ->prev_in_clause;
    }
    
    clause->size -= 1;
    
    assert(bit_vector_contains(clause->scopes, occ->var->scope->scope_id));
    
    // update scopes
    bool remove = true;
    for (Occ* clause_occ = clause->first_occ; clause_occ != NULL; clause_occ = clause_occ->next_in_clause) {
        if (clause_occ->var->scope->scope_id == occ->var->scope->scope_id) {
            // found another occurence with same scope_id
            //assert(clause_occ->var->scope == occ->var->scope);
            remove = false;
            break;
        }
    }
    if (remove) {
        bit_vector_remove(clause->scopes, occ->var->scope->scope_id);
    }
    assert(clause->size == 0 || bit_vector_min(clause->scopes) != BIT_VECTOR_NO_ENTRY);
    
    // check for further propagations
    if (matrix->phase == PHASE_BUILDING || matrix->phase == PHASE_PREPROCESSING) {
        check_propagation_from_clause(matrix, clause);
        reduce_clause_universally(matrix, clause);
    }
}

static void remove_occurrence_from_occurrence_list(Matrix* matrix, Occ* occ) {
    if (!occ->neg) {
        logging_debug("removing clause %d from pos occurrence list of variable %d\n", occ->clause->clause_id, occ->var->var_id);
        assert(occ->var->first_pos != NULL);
        assert(occ->var->last_pos != NULL);
        if(occ->var->first_pos == occ && occ->var->last_pos == occ) {
            // is the only occurrence
            occ->var->last_pos  = NULL;
            occ->var->first_pos = NULL;
        } else if (occ->var->first_pos == occ) {
            // is the first occurrence, but not the last
            occ->var->first_pos = occ->next_in_occ;
            occ->next_in_occ->prev_in_occ = NULL;
        } else if(occ->var->last_pos == occ) {
            // is the last occurrence, but not the first
            occ->var->last_pos = occ->prev_in_occ;
            occ->prev_in_occ->next_in_occ = NULL;
        } else { // is a middle node
            occ->prev_in_occ->next_in_occ = occ->next_in_occ;
            occ->next_in_occ->prev_in_occ = occ->prev_in_occ;
        }
        assert(occ->var->num_occ_pos > 0);
        occ->var->num_occ_pos -= 1;
    } else { // o->neg
        logging_debug("removing clause %d from neg occurrence list of variable %d\n", occ->clause->clause_id, occ->var->var_id);
        assert(occ->var->first_neg != NULL);
        assert(occ->var->last_neg != NULL);
        if(occ->var->first_neg == occ && occ->var->last_neg == occ) {
            // is the only occurrence
            occ->var->last_neg  = NULL;
            occ->var->first_neg = NULL;
        } else if (occ->var->first_neg == occ) {
            // is the first occurrence, but not the last
            occ->var->first_neg = occ->next_in_occ;
            occ->next_in_occ->prev_in_occ = NULL;
        } else if(occ->var->last_neg == occ) {
            // is the last occurrence, but not the first
            occ->var->last_neg = occ->prev_in_occ;
            occ->prev_in_occ->next_in_occ = NULL;
        } else { // is a middle node
            occ->prev_in_occ->next_in_occ = occ->next_in_occ;
            occ->next_in_occ->prev_in_occ = occ->prev_in_occ;
        }
        assert(occ->var->num_occ_neg > 0);
        occ->var->num_occ_neg -= 1;
    }
    
    // check if occurrence list of variable became empty
    propagate_var_with_empty_occurrence_list(matrix, occ->var);
}

static void reduce_clause_universally(Matrix* matrix, Clause* clause) {
    var_t max_scope = bit_vector_max(clause->scopes);
    if (max_scope % 2 == 0) {
        // max scope is existential
        return;
    }
    
    Occ* occ = clause->first_occ;
    while (occ != NULL) {
        if (matrix->conflict) {
            return;
        }
        if (occ->var->scope->scope_id == max_scope) {
            assert(occ->var->scope->qtype == QUANTIFIER_UNIVERSAL);
            Occ* remove = occ;
            occ = occ->next_in_clause;
            
            // encode precondition for universal reduction
            // format: [value] [literals]* 0
            // example: 1 -2 4 0  corresponding to clause (-2 4 -u)
            int_queue_push(&remove->var->reductions, remove->neg ? 1 : -1);
            for (Occ* pre = clause->first_occ; pre != NULL; pre = pre->next_in_clause) {
                if (pre == remove) {
                    continue;
                }
                int_queue_push(&remove->var->reductions, matrix_get_lit(pre));
            }
            int_queue_push(&remove->var->reductions, 0);
            
            remove_occurrence_from_clause(matrix, remove);
            remove_occurrence_from_occurrence_list(matrix, remove);
            free(remove);
            
            max_scope = bit_vector_max(clause->scopes);
            if (max_scope % 2 == 0) {
                // max scope is existential
                return;
            }
        } else {
            occ = occ->next_in_clause;
        }
    }
}

static void free_scope(Scope* scope) {
    assert(vector_count(scope->vars) == 0);
    vector_free(scope->vars);
    free(scope->next);
    free(scope);
}

static void remove_scope(Matrix* matrix, Scope* scope) {
    if (scope == matrix->top_level) {
        // Do not remove top-level scope
        return;
    }
    logging_info("scope %d became empty\n", scope->scope_id);
    assert(vector_count(scope->vars) == 0);
    Scope* previous = scope->prev;
    assert(previous != NULL);
    
    unlink_scope(previous, scope);
    
    // Merge previous and next scopes
    for (size_t i = 0; i < scope->num_next; i++) {
        Scope* next = scope->next[i];
        assert(next->qtype == previous->qtype);
        for (size_t j = 0; j < vector_count(next->vars); j++) {
            Var* var = vector_get(next->vars, j);
            vector_add(previous->vars, var);
            var->scope = previous;
        }
        vector_reset(next->vars);
        for (size_t j = 0; j < next->num_next; j++) {
            Scope* next_next = next->next[j];
            link_scope(previous, next_next);
        }
        for (Clause* clause = matrix->first_clause; clause != NULL; clause = clause->next) {
            assert(!bit_vector_contains(clause->scopes, scope->scope_id));
            if (bit_vector_contains(clause->scopes, next->scope_id)) {
                bit_vector_remove(clause->scopes, next->scope_id);
                bit_vector_add(clause->scopes, previous->scope_id);
            }
        }
     
        free_scope(next);
    }
    
    scope->scope_id = -1;
    free_scope(scope);
}

static void remove_variable(Matrix* matrix, Var* var) {
    Scope* scope = var->scope;
    assert(vector_contains(scope->vars, var));
    vector_remove(scope->vars, var);
    var->removed = true;
    
    assert(var->num_occ_neg + var->num_occ_pos == 0);
    
    if (vector_count(scope->vars) == 0) {
        remove_scope(matrix, scope);
    }
}

void matrix_preprocess(Matrix* matrix) {
    
    matrix->phase = PHASE_PREPROCESSING;

    assert(matrix_check(matrix));
    
    // check all variables for empty occurrence lists
    for (size_t i = 1; i <= matrix->max_vars; i++) {
        Var* var = matrix->vars[i];
        if (var == NULL) {
            continue;
        }
        assert(var->value == 0);
        propagate_var_with_empty_occurrence_list(matrix, var);
    }
    
    // Universal reductions:
    for (Clause* clause = matrix->first_clause; clause != NULL; clause = clause->next) {
        reduce_clause_universally(matrix, clause);
    }
    
    while (!int_queue_is_empty(&matrix->propagation_queue)) {
        lit_t propagation_lit = int_queue_pop(&matrix->propagation_queue);
        var_t propagation_var = lit_to_var(propagation_lit);
        logging_debug("propagating literal %d\n", propagation_lit);
        
        Var* var = matrix->vars[propagation_var];
        assert(var != NULL);
        
        if (var->removed) {
            continue;
        }
        
        if ((var->value == 1 && propagation_lit < 0) || (var->value == -1 && propagation_lit > 0)) {
            logging_info("propagation conflict for variable %d\n", propagation_var);
            matrix->conflict = true;
            continue;
            return;
        }
        int propagated_value = propagation_lit > 0 ? 1 : -1;
        assert(var->value == 0);
        var->value = propagated_value;
        
        // Iterate over occurrence list of propagated variable
        
        // all clauses where literal appeared with the same polarity are removed
        // (as removing the clause also removes the occurrence we iterate through, we have to remember the next list entry)
        logging_debug("removing satisifed claused due to propagation %d\n", propagation_lit);
        
        Occ* same_value_occs = (propagation_lit > 0) ? var->first_pos : var->first_neg;
        while (same_value_occs != NULL) {
            Occ* next_occ = same_value_occs->next_in_occ;
            remove_clause(matrix, same_value_occs->clause);
            same_value_occs = next_occ;
        }
        
        // remove literal from clauses where it appeared with different polarity
        logging_debug("removing %d from clauses due to propagation\n", var->var_id);
        Occ* diff_value_occs = (propagation_lit > 0) ? var->first_neg : var->first_pos;
        while (diff_value_occs) {
            Occ* next_occ = diff_value_occs->next_in_occ;
            assert(diff_value_occs->var->scope->scope_id == var->scope->scope_id);
            remove_occurrence_from_clause(matrix, diff_value_occs);
            remove_occurrence_from_occurrence_list(matrix, diff_value_occs);
            free(diff_value_occs);
            diff_value_occs = next_occ;
        }
        
        assert(var->num_occ_pos == 0);
        assert(var->num_occ_neg == 0);
        
        // Remove variable
        remove_variable(matrix, var);
        
        if (matrix->conflict) {
            return;
        }
    }
    
    assert(matrix_check(matrix));
}

static void union_over_connecting_sets(Matrix* matrix, Scope* scope, var_t* partitions) {
    for (Clause* clause = matrix->first_clause; clause != NULL; clause = clause->next) {
        bool has_connection = false;
        var_t connection = 0;
        
        for (Occ* occ = clause->first_occ; occ != NULL; occ = occ->next_in_clause) {
            if (occ->var->scope->qtype == QUANTIFIER_UNIVERSAL) {
                continue;
            }
            if (occ->var->scope->scope_id < scope->scope_id) {
                continue;
            }
            var_t var = occ->var->var_id;
            assert(var > 0);
            
            // Check whether this variable connects some variable sets
            while (partitions[partitions[var]] != partitions[var]) { // compactify
                partitions[var] = partitions[partitions[var]];
            }
            if (!has_connection) {
                connection = partitions[var];
                has_connection = true;
                continue;
            }
            assert(has_connection);
            if (connection < partitions[var]) {
                partitions[partitions[var]] = connection;
                partitions[var] = connection;
            }
            if (connection > partitions[var]) {
                partitions[connection] = partitions[var];
                connection = partitions[var];
            }
        }
    }
    
    for (size_t i = 1; i <= matrix->orig_max_vars; i++) {
        do {
            partitions[i] = partitions[partitions[i]];
        } while (partitions[i] >= i && partitions[i] != partitions[partitions[i]]);
    }
}

static void move_variable(Matrix* matrix, Var* var, Scope* from, Scope* to) {
    assert(matrix);
    assert(vector_contains(from->vars, var));
    vector_remove(from->vars, var);
    vector_add(to->vars, var);
    var->scope = to;
    
    /*if (vector_count(from->vars) == 0) {
        //remove_scope(matrix, from);
        
        // unlink scope but keep it allocated
        assert(from->num_next == 0);
        unlink_scope(from->prev, from);
    }*/
}

/**
 * change the order of clauses such that every partition corresponds to a
 * consecutive list of clauses.
 */
static void sort_clauses_by_partitions(Matrix* matrix, var_t* partitions, var_t pivot, Scope* current) {
    
    // keep the block of clauses that correspond to inner partitions
    for (size_t i = 0; i < current->num_next; i++) {
        Scope* next = current->next[i];
        if (next->first_clause == NULL) {
            assert(next->last_clause == NULL);
            fixme("empty scope during partitioning\n");
            continue;
        }
        assert(next->first_clause != NULL);
        
        unlink_clause_block(matrix, next->first_clause, next->last_clause);
        
        if (current->last_clause == NULL) {
            assert(current->first_clause == NULL);
            current->first_clause = next->first_clause;
            current->last_clause = next->last_clause;
        } else {
            insert_after(matrix, current->last_clause, next->first_clause);
            current->last_clause = next->last_clause;
        }
        assert(current->last_clause->next == NULL);
    }
    
    Clause* clause = matrix->first_clause;
    while (clause != NULL) {
        Clause* next = clause->next;
        
        for (Occ* occ = clause->first_occ; occ != NULL; occ = occ->next_in_clause) {
            var_t var = lit_to_var(matrix_get_lit(occ));
            if (occ->var->scope->qtype == QUANTIFIER_UNIVERSAL) {
                continue;
            }
            assert(var <= matrix->orig_max_vars);
            if (partitions[var] != pivot) {
                continue;
            }
            // clause corresponds to current partition
            // move it to the end of current->last_clause
            
            if (bit_vector_max(clause->scopes) < current->scope_id) {
                //assert(current->num_next == 0);
                assert(false);
                //clause->max_scope = current;
#ifndef NDEBUG
                matrix_check_clause(matrix, clause);
#endif
            }
            
            unlink_clause(matrix, clause);
            
            if (current->last_clause == NULL) {
                assert(current->first_clause == NULL);
                current->first_clause = current->last_clause = clause;
            } else {
                insert_after(matrix, current->last_clause, clause);
                current->last_clause = clause;
            }
            assert(current->last_clause->next == NULL);
            break;
        }
        
        clause = next;
    }
}

/**
 * After partitioning the matrix according to scope aprtitioning,
 * we combine the clauses to a consecutive matrix
 */
static void fix_clauses(Matrix* matrix, vector* scopes) {
    for (size_t i = 0; i < vector_count(scopes); i++) {
        Scope* scope = vector_get(scopes, i);
        assert(scope->first_clause != NULL);
        if (matrix->last_clause != NULL) {
            matrix->last_clause->next = scope->first_clause;
            scope->first_clause->prev = matrix->last_clause;
            matrix->last_clause = scope->last_clause;
        } else {
            matrix->first_clause = scope->first_clause;
            matrix->last_clause = scope->last_clause;
        }
    }
}

static void partition_scopes(Matrix* matrix, Scope* scope, var_t* partitions, vector* scopes) {
    size_t num_partitions = 0;
    map* groups = map_init();
    for (size_t i = 1; i <= matrix->max_vars; i++) {
        Var* var = matrix->vars[i];
        if (var == NULL) {
            continue;
        }
        assert(var->var_id == i);
        
        if (var->scope->scope_id < scope->scope_id) {
            continue;
        }
        if (var->scope->qtype != QUANTIFIER_EXISTENTIAL) {
            continue;
        }
        if (var->removed) {
            continue;
        }
        
        do {
            partitions[i] = partitions[partitions[i]];
        } while (partitions[i] >= i && partitions[i] != partitions[partitions[i]]);
        
        //for (size_t i = 1; i <= matrix->orig_max_vars; i++) {
        //    printf("%d -> %d\n", i, partitions[i]);
        //}
        
        logging_debug("variable %d is in partition %d\n", i, partitions[i]);
        
        if (partitions[i] == i) {
            var_t group_id = i;
            num_partitions++;
            assert(scope->qtype == QUANTIFIER_EXISTENTIAL);
            Scope* new_scope = create_scope(matrix, scope->qtype);
            new_scope->scope_id = scope->scope_id;  // copy scope_id
            new_scope->group = group_id;
            link_scope(scope->prev, new_scope);
            size_t j = 0;
            while (j < scope->num_next) {
                Scope* next = scope->next[j];
                assert(next->group != 0);
                if (partitions[next->group] == group_id) {
                    unlink_scope(scope, next);
                    link_scope(new_scope, next);
                } else {
                    j++;
                }
            }
            if (var->scope == scope) {
                move_variable(matrix, var, scope, new_scope);
            }
            map_add(groups, i, new_scope);
            vector_add(scopes, new_scope);
            sort_clauses_by_partitions(matrix, partitions, group_id, new_scope);
        } else {
            assert(partitions[i] < i);
            assert(num_partitions > 0);
            Scope* new_scope = map_get(groups, partitions[i]);
            assert(new_scope != NULL);
            if (var->scope == scope) {
                move_variable(matrix, var, scope, new_scope);
            }
        }
    }
    map_free(groups);
    logging_info("detected %zu partitions\n", num_partitions);
    
    //remove_scope(matrix, scope);
    // unlink scope but keep it allocated
    assert(scope->num_next == 0);
    unlink_scope(scope->prev, scope);
    
    if (scope == matrix->top_level) {
        assert(num_partitions <= 1);
    }
}

static Var* copy_var(Matrix* matrix, Var* to_copy, Scope* scope) {
    Var* copy = matrix_new_var(matrix, scope, matrix->max_vars + 1);
    copy->orig_var_id = to_copy->orig_var_id;
    return copy;
}

static void split_universal(Matrix* matrix, Scope* scope) {
    assert(scope->num_next != 0);
    if (scope->num_next == 1) {
        Scope* next = scope->next[0];
        scope->group = next->group;
        scope->first_clause = next->first_clause;
        scope->last_clause = next->last_clause;
        return;
    }
    
    assert(scope->prev != NULL);
    
    while (scope->num_next > 0) {
        Scope* next = scope->next[0];
        unlink_scope(scope, next);
        Scope* copy = create_scope(matrix, QUANTIFIER_UNIVERSAL);
        copy->scope_id = scope->scope_id;
        copy->group = next->group;
        assert(copy->group != 0);
        link_scope(copy, next);
        link_scope(scope->prev, copy);
        
        // copy variables
        map* var_renaming = map_init();
        for (size_t i = 0; i < vector_count(scope->vars); i++) {
            Var* var = vector_get(scope->vars, i);
            Var* var_copy = copy_var(matrix, var, copy);
            map_add(var_renaming, var->var_id, var_copy);
        }
        
        copy->first_clause = next->first_clause;
        copy->last_clause = next->last_clause;
        
        // iterate over clauses and rename universal variables
        for (Clause* clause = copy->first_clause; clause != NULL; clause = clause->next) {
            Occ* occ = clause->first_occ;
            while (occ != NULL) {
                // we modify the clause, so we have to remember next occurrence
                Occ* next_occ = occ->next_in_clause;
                var_t var = lit_to_var(matrix_get_lit(occ));
                Var* var_copy = map_get(var_renaming, var);
                if (var_copy == NULL) {
                    occ = next_occ;
                    continue;
                }
                occ->var = var_copy;
                
                // make sure that clause is still sorted
                remove_occurrence_from_clause(matrix, occ);
                add_occurrence_to_clause(clause, occ);
                
                occ = next_occ;
            }
            if (clause == copy->last_clause) {
                break;
            }
        }
        map_free(var_renaming);
    }
    
    //remove_scope(matrix, scope);
    // unlink scope but keep it allocated
    assert(scope->num_next == 0);
    unlink_scope(scope->prev, scope);
}

static void unprenex_by_miniscoping_recursive(Matrix* matrix, Scope* scope, var_t* partitions) {
    // we descend to the lowest level quantifier and build the partitioning bottom-up
    if (scope->num_next != 0) {
        // formulas can only be built with prenex quantification
        assert(scope->num_next == 1);
        unprenex_by_miniscoping_recursive(matrix, scope->next[0], partitions);
    }
    if (vector_count(scope->vars) == 0) {
        return;
    }
    if (scope == matrix->top_level) {
        fixme("partitioning on top-level not supported\n");
        return;
    }
    
    if (scope->qtype == QUANTIFIER_EXISTENTIAL) {
        union_over_connecting_sets(matrix, scope, partitions);
        vector* scopes = vector_init();
        partition_scopes(matrix, scope, partitions, scopes);
        fix_clauses(matrix, scopes);
        vector_free(scopes);
    } else if (scope->qtype == QUANTIFIER_UNIVERSAL) {
        split_universal(matrix, scope);
    }
}

static void rename_clauses(Matrix* matrix) {
    var_t clause_id = 0;
    for (Clause* clause = matrix->first_clause; clause != NULL; clause = clause->next) {
        matrix->clauses[clause_id] = clause;
        clause->clause_id = clause_id++;
    }
    for (var_t removed_clause = clause_id; removed_clause < matrix->num_clauses; removed_clause++) {
        matrix->clauses[removed_clause] = NULL;
    }
}

void matrix_unprenex_by_miniscoping(Matrix* matrix) {
    matrix->phase = PHASE_PARTITIONING;
    var_t* partitions = malloc((matrix->max_vars + 1) * sizeof(var_t));
    for (var_t i = 0; i <= matrix->max_vars; i++) {
        partitions[i] = i;
    }
    matrix->orig_max_vars = matrix->max_vars;
    unprenex_by_miniscoping_recursive(matrix, matrix->top_level, partitions);
    free(partitions);
    
    //assert(matrix->first_clause == matrix->top_level->first_clause);
    //assert(matrix->last_clause == matrix->top_level->last_clause);
    
    rename_clauses(matrix);
    
#ifndef NDEBUG
    matrix_check_partitioning(matrix);
#endif
}

