//
//  matrix.h
//  caqe
//
//  Copyright (c) 2015 - 2016, Leander Tentrup, Saarland University
//                2015, Markus Rabe, University of California, Berkeley
//
//  Licensed under Apache License v2.0, see LICENSE.txt for information
//

#ifndef __caqe__matrix__
#define __caqe__matrix__

#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>

#include "config.h"
#include "map.h"
#include "vector.h"
#include "queue.h"
#include "bit_vector.h"


typedef int32_t lit_t;
typedef uint32_t var_t;

typedef enum {
    QUANTIFIER_EXISTENTIAL,
    QUANTIFIER_UNIVERSAL
} quant_t;

typedef enum {
    QBF_RESULT_SAT     = 10,
    QBF_RESULT_UNSAT   = 20,
    QBF_RESULT_UNKNOWN = 30
} qbf_res;

typedef enum {
    PHASE_BUILDING,
    PHASE_PREPROCESSING,
    PHASE_PARTITIONING
} phase_t;

typedef struct Clause Clause;
typedef struct Var Var;
typedef struct Occ Occ;
typedef struct Scope Scope;
typedef struct Matrix Matrix;

struct Clause {
    var_t clause_id; // corresponds to T variable num
    var_t orig_clause_id;
    
    Clause* next;
    Clause* prev;
    
    size_t size; // number of variables
    Occ* first_occ;
    Occ* last_occ;
    
    bit_vector* scopes;
    
    int is_in_entry; // used for minimization
    bool satisified;
};

/**
 * an occurrence of a variable in a clause
 */
struct Occ {
    bool neg;
    Var* var;
    Clause* clause;
    Occ* next_in_clause;
    Occ* prev_in_clause;
    Occ* next_in_occ; // occurrence list
    Occ* prev_in_occ; // occurrence list
};

struct Var {
    var_t var_id;
    var_t orig_var_id;
    
    size_t num_occ_pos;
    Occ* first_pos;
    Occ* last_pos;
    
    size_t num_occ_neg;
    Occ* first_neg;
    Occ* last_neg;
    
    Scope* scope;
    int value; // -1, 0, 1
    bool removed;
    quant_t orig_quant;
    
    int_queue reductions; // stores in which clauses the variable was universally reduced
};

struct Scope {
    var_t scope_id;
    quant_t qtype;
    Matrix* matrix;
    vector* vars;

    Scope* prev;
    Scope** next;
    size_t num_next;
    
    // partitions
    Clause* first_clause;
    Clause* last_clause;
    var_t group;
};

// This data structure fulfills several purposes:
// (1) Efficient manipulation operations
// (2) Minimization throug propagation, pure literal detection, and universal reduction
// (3) Maintain occurrence lists
struct Matrix {
    Var** vars;     // maps var ids to Var*
    size_t max_vars;
    size_t size_vars;

    Clause* first_clause;
    Clause* last_clause;
    
    Clause** clauses;  // maps clause ids to Clause*
    size_t num_clauses;
    size_t size_clauses;
    
    // scopes
    var_t last_scope_id;
    Scope* previous_scope;
    Scope* top_level;
    
    // propagation
    bool conflict;
    int_queue propagation_queue;
    phase_t phase;
    
    // partitioning
    size_t orig_max_vars;
};


// Constructors
Matrix* matrix_init(void);
void matrix_adjust(Matrix*, size_t num_vars, size_t num_clauses);
void matrix_free(Matrix*);

// Variables
Var* matrix_new_var(Matrix*, Scope*, var_t);
Var* matrix_get_var(Matrix*, var_t);

// Scopes
Scope* matrix_new_scope(Matrix*, quant_t);
Clause* matrix_iterator_start(const Matrix*, const Scope*);
Clause* matrix_iterator_end(const Matrix*, const Scope*);

// Clauses
Clause* matrix_new_clause(Matrix*);
Occ* matrix_add_lit_to_clause(Matrix*, Clause*, lit_t);
void matrix_close_clause(Matrix*, Clause*);

// Preprocessing
void matrix_preprocess(Matrix*);
void matrix_unprenex_by_miniscoping(Matrix*);

// Helper
static inline lit_t create_lit(var_t var, bool negated) { return negated ? -(lit_t)var : (lit_t)var; }
static inline lit_t create_lit_from_value(var_t var, int value) { return value < 0 ? -(lit_t)var : (lit_t)var; }
static inline lit_t matrix_get_lit(Occ* occ) { return create_lit(occ->var->var_id, occ->neg); }
static inline var_t lit_to_var(lit_t lit) { return lit < 0 ? (var_t)-lit : (var_t)lit; }
static inline quant_t matrix_negate_quantifier(quant_t q) { return q == QUANTIFIER_UNIVERSAL ? QUANTIFIER_EXISTENTIAL : QUANTIFIER_UNIVERSAL; }

#endif /* defined(__caqe__matrix__) */
