//
//  matrix_check.c
//  caqe
//
//  Copyright (c) 2016, Leander Tentrup, Saarland University
//
//  Licensed under Apache License v2.0, see LICENSE.txt for information
//

#include "matrix_check.h"

#include <assert.h>

static void check_variables(Matrix* matrix) {
    for (size_t i = 1; i <= matrix->max_vars; i++) {
        Var* var = matrix->vars[i];
        if (var == NULL) {
            continue;
        }
        assert(var->var_id == i);
        
        // check occurrence lists
        size_t num_occ_pos = 0;
        for (Occ* occ = var->first_pos; occ != NULL; occ = occ->next_in_occ) {
            num_occ_pos++;
        }
        assert(var->num_occ_pos == num_occ_pos);
        
        size_t num_occ_neg = 0;
        for (Occ* occ = var->first_neg; occ != NULL; occ = occ->next_in_occ) {
            num_occ_neg++;
        }
        assert(var->num_occ_neg == num_occ_neg);
        
        size_t num_occ = 0;
        for (Clause* clause = matrix->first_clause; clause != NULL; clause = clause->next) {
            for (Occ* occ = clause->first_occ; occ != NULL; occ = occ->next_in_clause) {
                var_t var_id = lit_to_var(matrix_get_lit(occ));
                if (var_id == var->var_id) {
                    num_occ++;
                }
            }
        }
        assert(num_occ == num_occ_neg + num_occ_pos);
    }
}

void matrix_check_clause(Matrix* matrix, Clause* clause) {
    assert(clause->clause_id < matrix->num_clauses);
    assert(matrix->num_clauses <= matrix->size_clauses);
    assert(matrix->clauses[clause->clause_id] == clause);
    assert(bit_vector_min(clause->scopes) != BIT_VECTOR_NO_ENTRY || clause->size == 0);
    
    // check that clause is sorted
    Occ* prev = clause->first_occ;
    if (prev == NULL) {
        // clause is empty
        return;
    }
    for (Occ* occ = prev->next_in_clause; occ != NULL; occ = occ->next_in_clause) {
        assert(lit_to_var(matrix_get_lit(prev)) < lit_to_var(matrix_get_lit(occ)));
        prev = occ;
    }
    
    // check min/max scope
    bit_vector* scopes = bit_vector_init(0, matrix->last_scope_id);
    for (Occ* occ = clause->first_occ; occ != NULL; occ = occ->next_in_clause) {
        assert(occ->var != NULL);
        assert(bit_vector_min(clause->scopes) <= occ->var->scope->scope_id);
        assert(occ->var->scope->scope_id <= bit_vector_max(clause->scopes));
        bit_vector_add(scopes, occ->var->scope->scope_id);
    }
    assert(bit_vector_min(scopes) == bit_vector_min(clause->scopes));
    assert(bit_vector_max(scopes) == bit_vector_max(clause->scopes));
    bit_vector_free(scopes);
}

static void check_clauses(Matrix* matrix) {
    for (Clause* clause = matrix->first_clause; clause != NULL; clause = clause->next) {
        matrix_check_clause(matrix, clause);
    }
    
    for (var_t i = 0; i < matrix->num_clauses; i++) {
        Clause* clause = matrix->clauses[i];
        if (clause == NULL) {
            continue;
        }
        assert(clause->clause_id == i);
        
        matrix_check_clause(matrix, clause);
    }
}

static void check_scopes_recursive(Matrix* matrix, Scope* scope) {
    for (size_t i = 0; i < scope->num_next; i++) {
        Scope* next = scope->next[i];
        check_scopes_recursive(matrix, next);
    }
    
    if (scope->first_clause != NULL) {
        assert(scope->last_clause != NULL);
        
        // check that clauses corresponding to scope are consecutive
        for (Clause* clause = scope->first_clause; clause != scope->last_clause; clause = clause->next) {
            assert(clause != NULL);
        }
    }
}

bool matrix_check(Matrix* matrix) {
    check_variables(matrix);
    check_clauses(matrix);
    return true;
}

void matrix_check_partitioning(Matrix* matrix) {
    check_scopes_recursive(matrix, matrix->top_level);
    check_clauses(matrix);
}

