//
//  qdimacs.c
//  caqe
//
//  Copyright (c) 2015 - 2016, Leander Tentrup, Saarland University
//                2015, Markus Rabe, University of California, Berkeley
//
//  Licensed under Apache License v2.0, see LICENSE.txt for information
//

#include "qdimacs.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "logging.h"

#define GUNZIP "gunzip -c %s 2>/dev/null"

int matrix_from_qdimacs(Matrix* matrix, FILE* file) {
    api_expect(file, "The file object is invalid.");
    
    char c;
    int lit = -1;
    bool neg = false;
    Scope* scope = NULL;
    Clause* clause = NULL;
    while ((c = getc(file)) != EOF) {
        switch (c) {
            case 'c':  // comment
                while (getc(file) != '\n');
                break;
                
            case 'p':  // header
                while (getc(file) != '\n');
                break;
                
            case 'e':  // quantifier
            case 'a':
                assert(scope == NULL);
                const quant_t qtype = (c == 'a') ? QUANTIFIER_UNIVERSAL : QUANTIFIER_EXISTENTIAL;
                scope = matrix_new_scope(matrix, qtype);
                logging_debug("%c ", c);
                break;
                
            case ' ':  // spaces
            case '\n':
                if (lit == -1) {
                    break;
                }
                if (scope == NULL && clause == NULL) {
                    clause = matrix_new_clause(matrix);
                }
                if (lit == 0) {
                    logging_debug("0\n");
                    if (clause != NULL) {
                        matrix_close_clause(matrix, clause);
                        clause = NULL;
                    } else {
                        assert(scope != NULL);
                        scope = NULL;
                    }
                } else {
                    assert(lit >= 0);
                    logging_debug("%d ", neg ? -lit : lit);
                    if (clause != NULL) {
                        matrix_add_lit_to_clause(matrix, clause, neg ? -lit : lit);
                    } else {
                        assert(scope != NULL);
                        api_expect(!neg, "negation is not allowed in scope");
                        matrix_new_var(matrix, scope, lit);
                    }
                }
                lit = -1;
                neg = false;
                break;
                
            case '-':
                assert(neg == false);
                neg = true;
                break;
                
            case '0':
            case '1':
            case '2':
            case '3':
            case '4':
            case '5':
            case '6':
            case '7':
            case '8':
            case '9':
                
                if (lit == -1) {
                    lit = 0;
                }
                lit = (lit * 10) + (c - '0');
                break;
                
            default:  // literals
                logging_error("Unknown character \"%c\"\n", c);
                return 1;
        }
    }
    
    if (clause != NULL) {
        matrix_close_clause(matrix, clause);
        logging_debug("0\n");
    }
    
    return 0;
}

static bool has_suffix(const char* string, const char* suffix) {
    if (strlen(suffix) > strlen(string)) {
        return false;
    }
    if (strncmp(string + (strlen(string) - strlen(suffix)), suffix, strlen(suffix)) == 0) {
        return true;
    }
    return false;
}

int qdimacs_open_and_read_qcir_file(Matrix* matrix, const char* file_name) {
    FILE* file;
    bool use_pclose;
    
    if (has_suffix(file_name, ".gz")) {
        use_pclose = true;
        
        char* cmd = malloc(strlen(file_name) + strlen(GUNZIP));
        sprintf(cmd, GUNZIP, file_name);
        file = popen(cmd, "r");
        free(cmd);
    } else {
        file = fopen(file_name, "r");
        use_pclose = false;
    }
    
    if (!file) {
        logging_fatal("File \"%s\" does not exist!\n", file_name);
        return -1;
    }
    
    int error = matrix_from_qdimacs(matrix, file);
    
    if (use_pclose) {
        pclose(file);
    } else {
        fclose(file);
    }
    
    return error;
}
