//
//  satsolver_minisat.cc
//  caqe
//
//  Copyright (c) 2015 - 2016, Leander Tentrup, Saarland University
//
//  Licensed under Apache License v2.0, see LICENSE.txt for information
//

extern "C" {
#include "satsolver.h"
}

#include <assert.h>
#include <stdbool.h>
#include "minisat/Solver.h"
#include "minisat/Vec.h"
#include <map>

using namespace Minisat;
using namespace std;

struct SATSolver {
    Solver* instance;
    vec<Lit> clause;
    vec<Lit> assumptions;
    map<int,Var> var_mapping;
    map<Var,int> var_reverse_mapping;
    int* failed_assumptions;
    bool last_sat_call;
};

Var var_from_int(SATSolver* solver, int var) {
    assert(var > 0);
    
    // lookup variable
    Var nvar;
    map<int,Var>::iterator elem = solver->var_mapping.find(var);
    if (elem == solver->var_mapping.end()) {
        // setting u_pol to l_False seems to favor negative assignments
        nvar = solver->instance->newVar(l_False);
        solver->var_mapping[var] = nvar;
        solver->var_reverse_mapping[nvar] = var;
    } else {
        nvar = elem->second;
    }
    
    return nvar;
}

Lit lit_from_int(SATSolver* solver, int lit) {
    bool neg = lit < 0;
    int var = neg ? -lit : lit;
    
    return mkLit(var_from_int(solver, var), neg);
}

int lit_to_int(SATSolver* solver, Lit lit) {
    bool neg = sign(lit);
    int v = var(lit);
    return neg ? -solver->var_reverse_mapping[v] : solver->var_reverse_mapping[v];
}

SATSolver* satsolver_init() {
    SATSolver* solver = new SATSolver;
    solver->instance = new Solver();
    solver->failed_assumptions = NULL;
    solver->last_sat_call = false;
    
    return solver;
}

void satsolver_free(SATSolver* solver) {
    delete solver->instance;
    delete solver;
}

void satsolver_adjust(SATSolver* solver, int variables) {}

void satsolver_save_original_clauses(SATSolver* solver) {}

void satsolver_reset(SATSolver* solver) {
    delete solver->instance;
    free(solver);
}

void satsolver_add(SATSolver* solver, int lit) {
    if (lit == 0) {
        solver->instance->addClause(solver->clause);
        solver->clause.clear();
    } else {
        Lit l = lit_from_int(solver, lit);
        solver->clause.push(l);
    }
}

void satsolver_assume(SATSolver* solver, int lit) {
    if (solver->last_sat_call) {
        solver->assumptions.clear();
        solver->last_sat_call = false;
    }
    solver->assumptions.push(lit_from_int(solver, lit));
}

void satsolver_clear_assumptions(SATSolver* solver) {
    solver->assumptions.clear();
}

sat_res satsolver_sat(SATSolver* solver) {
    if (solver->last_sat_call) {
        solver->assumptions.clear();
        solver->last_sat_call = false;
    }
    sat_res res = solver->instance->solve(solver->assumptions) ? SATSOLVER_RESULT_SAT : SATSOLVER_RESULT_UNSAT;
    solver->last_sat_call = true;
    return res;
}

int satsolver_deref(SATSolver* solver, int lit) {
    
    /* TODO: workaround: if the variable is not contained in the sat solver, return undefined */
    int var = lit < 0 ? -lit : lit;
    map<int,Var>::iterator elem = solver->var_mapping.find(var);
    if (elem == solver->var_mapping.end()) {
        return 0;
    }
    
    lbool res = solver->instance->modelValue(lit_from_int(solver, lit));
    if (res == l_True) {
        return 1;
    } else if (res == l_False) {
        return -1;
    } else if (res == l_Undef) {
        return 0;
    }
    assert(false);
    return 0;
}

int satsolver_deref_partial(SATSolver* solver, int lit) {
    return satsolver_deref(solver, lit);
}

const int* satsolver_failed_assumptions(SATSolver* solver) {
    LSet& conflict = solver->instance->conflict;
    free(solver->failed_assumptions);
    solver->failed_assumptions = (int *)malloc(sizeof(int) * (conflict.size() + 1));
    
    int i = 0;
    for (int j = 0; j < solver->assumptions.size(); j++) {
        Lit elem = solver->assumptions[j];
        Lit pure = ~elem;
        if (conflict.has(pure)) {
            solver->failed_assumptions[i++] = lit_to_int(solver, elem);
        }
    }

    /*for (i = 0; i < conflict.size(); i++) {
        Lit failed = conflict[i];
        //solver->failed_assumptions[i] = lit_to_int(solver, failed);
    }*/
    
    assert(i == conflict.size());
    solver->failed_assumptions[i] = 0;
    return solver->failed_assumptions;
}

double satsolver_seconds(SATSolver* solver) {
    return 0;
}


void satsolver_set_global_default_phase(SATSolver* s, int phase) {
    return; // TODO
}

void satsolver_set_default_phase_lit (SATSolver* solver, int lit, int phase) {
    lbool lphase = l_False;
    if (phase > 0) {
        lphase = l_True;
    }
    solver->instance->setPolarity(var_from_int(solver, lit), lphase);
}

void satsolver_print(SATSolver* s) {
    return; // TODO
}

const char* satsolver_get_name() {
    return "MiniSat";
}

const char* satsolver_get_version() {
    return "2.2.0";
}

const char* satsolver_copyright() {
    return "Copyright 2003 - 2006 Niklas Een, 2007 - 2012 Niklas Sorensson";
}


