//
//  solver.c
//  caqe
//
//  Copyright (c) 2015 - 2016, Leander Tentrup, Saarland University
//                2015, Markus Rabe, University of California, Berkeley
//
//  Licensed under Apache License v2.0, see LICENSE.txt for information
//

#include "solver.h"

#include <assert.h>

#include "certification.h"
#include "clausal_abstraction.h"
#include "config.h"
#include "logging.h"
#include "matrix.h"
#include "matrix_print.h"

#define STRINGIFY_(x) #x
#define STRINGIFY(x) STRINGIFY_(x)

#ifdef CAQE_RELEASE_VERSION
    #define VERSION_STRING STRINGIFY(CAQE_MAJOR_VERSION) "." STRINGIFY(CAQE_MINOR_VERSION)
#else
    #define VERSION_STRING STRINGIFY(CAQE_MAJOR_VERSION) "." STRINGIFY(CAQE_MINOR_VERSION) "-dev"
#endif

#ifdef PARALLEL_SOLVING
static void* solve_sub_untyped(void* data);
#endif


static ClausalAbstraction* abstraction_init(Solver* solver, Scope* scope, ClausalAbstraction* prev) {
    ClausalAbstraction* abstraction = clausal_abstractions_init(solver, scope, prev);
    for (size_t i = 0; i < scope->num_next; i++) {
        abstraction->next[i] = abstraction_init(solver, scope->next[i], abstraction);
        
#ifdef PARALLEL_SOLVING
        if (scope->num_next > 1) {
            semaphore_init(&abstraction->next[i]->has_entry, 0);
            pthread_create(&abstraction->next[i]->thread, NULL, solve_sub_untyped, abstraction->next[i]);
        }
#endif
    }
#ifndef NDEBUG
    // check consistency of interface literals
    // for every b_lit in abstraction, there is a corresponding t_lit in one of its inner abstractions
    for (size_t i = 0; i < int_vector_count(abstraction->b_lits); i++) {
        const lit_t b_lit = int_vector_get(abstraction->b_lits, i);
        const lit_t t_lit = scope->qtype == QUANTIFIER_UNIVERSAL ? b_lit : b_lit_to_t_lit(solver, b_lit);
        size_t num_contained = 0;
        for (size_t j = 0; j < scope->num_next; j++) {
            ClausalAbstraction* next = abstraction->next[j];
            if (int_vector_contains_sorted(next->t_lits, t_lit)) {
                num_contained++;
            }
        }
        //const var_t clause_id = t_lit_to_clause_id(solver, t_lit);
        //Clause* clause = solver->matrix->clauses[clause_id];
        //matrix_print_clause(solver->matrix, clause);
        assert(num_contained == 1);
    }
#endif
    
    return abstraction;
}


static void assume_t_literals(Solver* solver, ClausalAbstraction* abstraction, bool negation) {
    
    const quant_t qtype = negation ? matrix_negate_quantifier(abstraction->scope->qtype) : abstraction->scope->qtype;
    SATSolver* sat = negation ? abstraction->negation : abstraction->sat;
    
    logging_debug("assume: ");
    for (size_t i = 0; i < int_vector_count(abstraction->t_lits); i++) {
        lit_t t_lit = int_vector_get(abstraction->t_lits, i);
        const var_t clause_id = t_lit_to_clause_id(solver, t_lit);
        if (!int_vector_contains_sorted(abstraction->entry, t_lit)) {
            t_lit = -t_lit;
        }
        
        if (negation) {
            t_lit = -t_lit;
        }
        
        logging_debug(t_lit > 0 ? "t%d " : "-t%d ", clause_id);
        
        if (!negation && t_lit > 0 && qtype == QUANTIFIER_UNIVERSAL) {
            // assume t-lit completely for existential quantifier
            // and only negative assignments for universal quantifier
            continue;
        }

        satsolver_assume(sat, t_lit);
    }
    logging_debug("\n");
}

static void add_t_lit_to_sub_entry(Solver* solver, ClausalAbstraction* abstraction, lit_t t_lit) {
    assert(solver);
    size_t num_added = 0;
    for (size_t j = 0; j < abstraction->scope->num_next; j++) {
        ClausalAbstraction* next = abstraction->next[j];
        assert(next != NULL);
        if (int_vector_contains_sorted(next->t_lits, t_lit)) {
            int_vector_add(next->entry, t_lit);
            num_added++;
        }
    }
    assert(num_added == 1);
}

static void get_assignment(Solver* solver, ClausalAbstraction* abstraction) {
    assert(solver);
    logging_debug("assignment: ");
    for (size_t i = 0; i < vector_count(abstraction->scope->vars); i++) {
        Var* var = vector_get(abstraction->scope->vars, i);
        int value = satsolver_deref(abstraction->sat, var->var_id);
        if (value != 0) {
            logging_debug("%d ", create_lit_from_value(var->var_id, value));
        }
        var->value = value;
    }
    logging_debug("\n");
}

static bool clause_is_satisified(ClausalAbstraction* abstraction, const Clause* clause) {
    for (Occ* occ = clause->first_occ; occ != NULL; occ = occ->next_in_clause) {
        if (occ->var->scope != abstraction->scope) {
            continue;
        }
        if (satsolver_deref(abstraction->sat, matrix_get_lit(occ)) > 0) {
            return true;
        }
        
        // alternative:
        //lit_t lit = matrix_get_lit(occ);
        //lit_t assignment = create_lit_from_value(occ->var->var_id, occ->var->value);
        //if (lit == assignment) {
        //    return true;
        //}
    }
    return false;
}

static void get_assumptions(Solver* solver, ClausalAbstraction* abstraction) {
    for (size_t i = 0; i < abstraction->scope->num_next; i++) {
        ClausalAbstraction* next = abstraction->next[i];
        assert(next != NULL);
        int_vector_reset(next->entry);
    }
    
    logging_debug("disabled: ");
    if (abstraction->scope->qtype == QUANTIFIER_EXISTENTIAL) {
        for (size_t i = 0; i < int_vector_count(abstraction->b_lits); i++) {
            const lit_t b_lit = int_vector_get(abstraction->b_lits, i);
            int value = satsolver_deref(abstraction->sat, b_lit);
            if (value >= 0) {
                continue;
            }
            
            const lit_t t_lit = b_lit_to_t_lit(solver, b_lit);
            
            const var_t clause_id = b_lit_to_clause_id(solver, b_lit);
            Clause* clause = solver->matrix->clauses[clause_id];
            assert(clause->clause_id == clause_id);
            bool satisfied = clause_is_satisified(abstraction, clause);

            if (satisfied || int_vector_contains_sorted(abstraction->entry, t_lit)) {
                continue;
            }
            
            // b-lit was disabled by SAT solver
            logging_debug("b%d ", clause_id);
            
            add_t_lit_to_sub_entry(solver, abstraction, t_lit);
        }
    } else {
        assert(abstraction->scope->qtype == QUANTIFIER_UNIVERSAL);
        for (size_t i = 0; i < int_vector_count(abstraction->b_lits); i++) {
            const lit_t t_lit = int_vector_get(abstraction->b_lits, i);
            int value = satsolver_deref(abstraction->sat, t_lit);
            if (value >= 0) {
                continue;
            }
            
            const var_t clause_id = t_lit_to_clause_id(solver, t_lit);
            Clause* clause = solver->matrix->clauses[clause_id];
            bool falsified = true;
            bool nonempty = false;
            for (Occ* occ = clause->first_occ; occ != NULL; occ = occ->next_in_clause) {
                if (occ->var->scope != abstraction->scope) {
                    continue;
                }
                nonempty = true;
                lit_t lit = matrix_get_lit(occ);
                lit_t assignment = create_lit_from_value(occ->var->var_id, occ->var->value);
                if (lit == assignment) {
                    falsified = false;
                    break;
                }
            }
            if (nonempty && falsified) {
                if (int_vector_contains_sorted(abstraction->t_lits, t_lit)) {
                    if (int_vector_contains_sorted(abstraction->entry, t_lit)) {
                        continue;
                    }
                } else {
                    continue;
                }
            }
            if (!nonempty) {
                assert(int_vector_contains_sorted(abstraction->t_lits, t_lit));
                if (int_vector_contains_sorted(abstraction->entry, t_lit)) {
                    continue;
                }
            }
            /*if (!nonempty && int_vector_contains_sorted(abstraction->entry, t_lit)) {
                continue;
            }*/
            
            // t-lit was disabled by SAT solver
            logging_debug("b%d ", clause_id);
            
            add_t_lit_to_sub_entry(solver, abstraction, t_lit);
        }
    }
    logging_debug("\n");
}

static void get_unsat_core(Solver* solver, ClausalAbstraction* abstraction) {
    int_vector_reset(abstraction->entry);
    const int* failed_assumptions = satsolver_failed_assumptions(abstraction->sat);
    logging_debug("unsat: ");
    for (size_t i = 0; failed_assumptions[i] != 0; i++) {
        const lit_t failed_t_lit = -failed_assumptions[i];
        assert(failed_t_lit > 0);
        const lit_t failed_clause = t_lit_to_clause_id(solver, failed_t_lit);
        int_vector_add(abstraction->entry, failed_t_lit);
        logging_debug("-t%d ", failed_clause);
    }
    logging_debug("\n");
}

static void append_to_min_entry(Solver* solver, ClausalAbstraction* abstraction, ClausalAbstraction* child) {
    assert(solver);
    for (size_t i = 0; i < int_vector_count(child->entry); i++) {
        const lit_t t_lit = int_vector_get(child->entry, i);
        assert(t_lit > 0);
        int_vector_add(abstraction->min_entry, t_lit);
        if (solver->options->assume_partial_results && abstraction->scope->qtype == QUANTIFIER_EXISTENTIAL) {
            satsolver_set_default_phase_lit(abstraction->sat, t_lit_to_b_lit(solver, t_lit), 1);
        }
    }
}

/**
 * Returns true, if the lits contained in lhs is a subset of the lits in rhs
 * (regarding the inner variables according to current scope).
 * Note that literals in clauses are sorted by their absolute value.
 */
static bool clause_is_subset(Clause* lhs, Clause* rhs, Scope* current_scope) {
    if (lhs == rhs) {
        return true;
    }
    Occ* rhs_occ = rhs->first_occ;
    Occ* lhs_occ = lhs->first_occ;
    
    // iterate over all literals in lhs and try to match it to the literals in rhs
    while (lhs_occ != NULL) {
        if (rhs_occ == NULL) {
            // remaining variables in lhs must be outer variables
            if (lhs_occ->var->scope->scope_id > current_scope->scope_id) {
                return false;
            }
            lhs_occ = lhs_occ->next_in_clause;
            continue;
        }
        
        if (lit_to_var(matrix_get_lit(lhs_occ)) == lit_to_var(matrix_get_lit(rhs_occ))) {
            if (matrix_get_lit(lhs_occ) != matrix_get_lit(rhs_occ)) {
                assert(matrix_get_lit(lhs_occ) == -matrix_get_lit(rhs_occ));
                return false;
            }
            // same literals, proceed with both
            lhs_occ = lhs_occ->next_in_clause;
            rhs_occ = rhs_occ->next_in_clause;
        } else if (lit_to_var(matrix_get_lit(rhs_occ)) < lit_to_var(matrix_get_lit(lhs_occ))) {
            // proceed with next lit in rhs
            rhs_occ = rhs_occ->next_in_clause;
        } else {
            // proceed with lhs if not outer variables
            if (lhs_occ->var->scope->scope_id > current_scope->scope_id) {
                return false;
            }
            lhs_occ = lhs_occ->next_in_clause;
        }
    }
    return true;
}

static void strong_unsat_refinement(Solver* solver, ClausalAbstraction* abstraction, ClausalAbstraction* child) {
    assert(abstraction->scope->qtype == QUANTIFIER_EXISTENTIAL);
    assert(solver->options->strong_unsat_refinement);
    
    // Strong unsat refinement: Exclude equivalent UNSAT cores
    
    int_vector_reset(abstraction->refinement);
    int_vector_reset(abstraction->conjunctive_lits);
    for (size_t i = 0; i < int_vector_count(child->entry); i++) {
        int failed_t_lit = int_vector_get(child->entry, i);
        assert(failed_t_lit > 0);
        const lit_t failed_clause = t_lit_to_clause_id(solver, failed_t_lit);
        const lit_t b_lit = t_lit_to_b_lit(solver, failed_t_lit);
        
        assert(int_vector_contains_sorted(abstraction->b_lits, b_lit));
        
        // Was the t-lit processed before?
        // TODO: disabling caching is not optimal (proof could save the auxilliary variables as well)
        void* entry = map_get(abstraction->t_lit_mapping, failed_t_lit);
        if (entry != NULL) {
            lit_t lit = (int)(intptr_t)entry;
            int_vector_add(abstraction->refinement, lit);
            continue;
        }
        
        // Get some random existential occurrence from clause, so we can use
        // the occurrence list to iterate over clauses
        Clause* clause = solver->matrix->clauses[failed_clause];
        int_vector_reset(abstraction->conjunctive_lits);
        for (Occ* occ = clause->first_occ; occ != NULL; occ = occ->next_in_clause) {
            Scope* var_scope = occ->var->scope;
            
            // Consider only existential variables that have a lower level
            if (var_scope->qtype == QUANTIFIER_UNIVERSAL || var_scope->scope_id <= abstraction->scope->scope_id) {
                continue;
            }
            
            // Iterate over occurrence list and add equivalent clauses
            Occ* var_occ = !occ->neg ? occ->var->first_pos : occ->var->first_neg;
            for (; var_occ != NULL; var_occ = var_occ->next_in_occ) {
                lit_t other_b_lit = clause_to_b_lit(solver, var_occ->clause);
                if (int_vector_contains_sorted(abstraction->conjunctive_lits, other_b_lit)) {
                    continue;
                }
                if (bit_vector_min(var_occ->clause->scopes) > abstraction->scope->scope_id) {
                    continue;
                }
                // clause is subset is probably more expensive that sorted find, that's why it is guarded here
                if (clause_is_subset(var_occ->clause, clause, abstraction->scope)) {
                    int_vector_add_sorted(abstraction->conjunctive_lits, other_b_lit);
                }
            }
        }
        
        assert(int_vector_count(abstraction->conjunctive_lits) > 0);
        
        //int_vector_print(conjunctive_lits);
        if (int_vector_count(abstraction->conjunctive_lits) == 1) {
            // do not need auxilliary variable
            lit_t other_b_lit = int_vector_get(abstraction->conjunctive_lits, 0);
            assert(int_vector_contains_sorted(abstraction->b_lits, other_b_lit));
            int_vector_add(abstraction->refinement, other_b_lit);
            map_add(abstraction->t_lit_mapping, failed_t_lit, (void*)(intptr_t)b_lit);
        } else {
            // Build the conjunction using an auxilliary variable
            lit_t aux_var = abstraction->aux_var++;
            int_vector_add(abstraction->refinement, aux_var);
            map_add(abstraction->t_lit_mapping, failed_t_lit, (void*)(intptr_t)aux_var);
            
            for (size_t j = 0; j < int_vector_count(abstraction->conjunctive_lits); j++) {
                lit_t other_b_lit = int_vector_get(abstraction->conjunctive_lits, j);
                assert(int_vector_contains_sorted(abstraction->b_lits, other_b_lit));
                
                satsolver_add(abstraction->sat, -aux_var);
                satsolver_add(abstraction->sat, other_b_lit);
                satsolver_add(abstraction->sat, 0);
            }
        }
    }
    
    for (size_t i = 0; i < int_vector_count(abstraction->refinement); i++) {
        lit_t lit = int_vector_get(abstraction->refinement, i);
        satsolver_add(abstraction->sat, lit);
        logging_debug("%d ", lit);
    }
    satsolver_add(abstraction->sat, 0);
    logging_debug("\n");
}

static void refine(Solver* solver, ClausalAbstraction* abstraction, ClausalAbstraction* child) {
    logging_debug("refine: ");
    if (abstraction->scope->qtype == QUANTIFIER_EXISTENTIAL && solver->options->strong_unsat_refinement) {
        strong_unsat_refinement(solver, abstraction, child);
        return;
    }
    for (size_t i = 0; i < int_vector_count(child->entry); i++) {
        const lit_t failed_t_lit = int_vector_get(child->entry, i);
        assert(failed_t_lit > 0);
        const lit_t failed_clause = t_lit_to_clause_id(solver, failed_t_lit);
        
        if (abstraction->scope->qtype == QUANTIFIER_EXISTENTIAL) {
            const lit_t b_lit = t_lit_to_b_lit(solver, failed_t_lit);
            
            if (!int_vector_contains_sorted(abstraction->b_lits, b_lit)) {
                assert(false);
                continue;
            }
            satsolver_add(abstraction->sat, b_lit);
            logging_debug("b%d ", failed_clause);
        } else {
            if (!int_vector_contains_sorted(abstraction->b_lits, failed_t_lit)) {
                assert(false);
                continue;
            }
            
            satsolver_add(abstraction->sat, failed_t_lit);
            logging_debug("t%d ", failed_clause);
        }
    }
    satsolver_add(abstraction->sat, 0);
    logging_debug("\n");
}

static void assume_assignment(Solver* solver, ClausalAbstraction* abstraction) {
    assert(solver);
    for (size_t i = 0; i < vector_count(abstraction->scope->vars); i++) {
        Var* var = vector_get(abstraction->scope->vars, i);
        int value = var->value;
        if (value != 0) {
            satsolver_assume(abstraction->negation, create_lit_from_value(var->var_id, value));
        }
    }
}

static void get_sat_assumptions(Solver* solver, ClausalAbstraction* abstraction) {
    int_vector_reset(abstraction->assumptions);
    
    if (abstraction->scope->num_next == 0) {
        return;
    }
    if (abstraction->scope->qtype == QUANTIFIER_UNIVERSAL) {
        return;
    }
    
    // We need to remember clauses for which a t_lit was set to true in entry,
    // but there is no corresponding b_lit, i.e., the inner quantifier do not
    // influence the clause.
    // We have to add those t_lits to the entry after minimization
    for (size_t i = 0; i < int_vector_count(abstraction->entry); i++) {
        // clauses with max_scope == current current scope need special treatment
        // when such a clause was satisifed by outer scope (t_lit in entry),
        // we have to add those t_lit's to the minimized entry
        
        const lit_t t_lit = int_vector_get(abstraction->entry, i);
        const lit_t b_lit = t_lit_to_b_lit(solver, t_lit);
        
        if (abstraction->scope->qtype == QUANTIFIER_UNIVERSAL) {
            // this situation cannot happen for universal quantified quantifiers
            // as it is guranteed that there is an existential quantifier below
            assert(int_vector_contains_sorted(abstraction->b_lits, t_lit));
            continue;
        }
        
        if (int_vector_contains_sorted(abstraction->b_lits, b_lit)) {
            continue;
        }
        
        const var_t clause_id = t_lit_to_clause_id(solver, t_lit);
        const Clause* clause = solver->matrix->clauses[clause_id];
        if (clause_is_satisified(abstraction, clause)) {
            continue;
        }
        
        int_vector_add(abstraction->assumptions, t_lit);
    }
}

static void dual_propagation(Solver* solver, ClausalAbstraction* abstraction) {
    
    assume_assignment(solver, abstraction);
    assume_t_literals(solver, abstraction, true);
    
    if (abstraction->scope->num_next != 0) {
        // Add UNSAT core as blocking clause
        for (size_t i = 0; i < int_vector_count(abstraction->min_entry); i++) {
            const lit_t failed_t_lit = int_vector_get(abstraction->min_entry, i);
            assert(failed_t_lit > 0);
            if (abstraction->scope->qtype == QUANTIFIER_UNIVERSAL) {
                const lit_t b_lit = t_lit_to_b_lit(solver, failed_t_lit);
                
                if (!int_vector_contains_sorted(abstraction->b_lits, failed_t_lit)) {
                    assert(false);
                    continue;
                }
                satsolver_add(abstraction->negation, b_lit);
            } else {
                const lit_t b_lit = t_lit_to_b_lit(solver, failed_t_lit);
                
                if (!int_vector_contains_sorted(abstraction->b_lits, b_lit)) {
                    assert(false);
                    continue;
                }
                satsolver_add(abstraction->negation, b_lit);
            }
        }
        if (abstraction->local_refinement_literals) {
            for (size_t i = 0; i < int_vector_count(abstraction->local_refinement_literals); i++) {
                const lit_t b_lit = int_vector_get(abstraction->local_refinement_literals, i);
                satsolver_add(abstraction->negation, b_lit);
            }
        }
        satsolver_add(abstraction->negation, 0);
    }
    
    sat_res result = satsolver_sat(abstraction->negation);
    assert(result == SATSOLVER_UNSATISFIABLE);
    
    // Reset variable assignments
    /*for (size_t i = 0; i < vector_count(abstraction->scope->vars); i++) {
        Var* var = vector_get(abstraction->scope->vars, i);
        var->value = 0;
    }*/
    
    int_vector_reset(abstraction->entry);
    const int* failed_assumptions = satsolver_failed_assumptions(abstraction->negation);
    logging_debug("min ");
    for (size_t i = 0; failed_assumptions[i] != 0; i++) {
        const lit_t failed_lit = failed_assumptions[i];  // may be variable or t-lit
        const var_t failed_var = lit_to_var(failed_lit);
        if (!int_vector_contains_sorted(abstraction->t_lits, failed_var)) {
            // variable assignments
            /*Var* var = solver->matrix->vars[failed_var];
            assert(var->scope == abstraction->scope);
            int value = create_lit_from_value(1, failed_lit);
            var->value = value;*/
            continue;
        }
        const lit_t failed_t_lit = -failed_lit;
        assert(failed_t_lit > 0);
        
        const var_t clause_id = t_lit_to_clause_id(solver, failed_t_lit);
        logging_debug("t%d ", clause_id);
        
        //Clause* clause = abstraction->scope->matrix->clauses[clause_id];
        //printf("\n");
        //matrix_print_clause(abstraction->scope->matrix, clause);
        
        int_vector_add(abstraction->entry, failed_t_lit);
        
#ifdef CERTIFICATION
        if (solver->options->certify) {
            certification_add_t_literal(&solver->certificate, abstraction, abstraction->scope->qtype, clause_id, t_lit_to_b_lit(solver, failed_t_lit), failed_t_lit);
        }
#endif
    }
    logging_debug("\n");
    
#ifdef CERTIFICATION
    if (solver->options->certify) {
        certification_append_function_case(&solver->certificate, abstraction);
    }
#endif
}

static void unsat_propagation(Solver* solver, ClausalAbstraction* abstraction) {
    assert(solver);
    int_vector_copy(abstraction->min_entry, abstraction->entry);
    for (size_t i = 0; i < int_vector_count(abstraction->assumptions); i++) {
        lit_t t_lit = int_vector_get(abstraction->assumptions, i);
        int_vector_add(abstraction->assumptions, t_lit);
    }
}

static qbf_res solve_recursive(Solver*, ClausalAbstraction*);

#ifdef PARALLEL_SOLVING
static qbf_res solve_sub(ClausalAbstraction* next) {
    Solver* solver = next->solver;
    ClausalAbstraction* abstraction = next->prev;
    const bool is_existential = abstraction->scope->qtype == QUANTIFIER_EXISTENTIAL;
    const qbf_res good_result = is_existential ? QBF_RESULT_SAT : QBF_RESULT_UNSAT;
    const qbf_res bad_result = is_existential ? QBF_RESULT_UNSAT : QBF_RESULT_SAT;
    
    qbf_res sub_result = solve_recursive(solver, next);
    pthread_mutex_lock(&abstraction->mutex);
    if (sub_result == good_result) {
        append_to_min_entry(solver, abstraction, next);
    } else {
        assert(sub_result == bad_result);
        refine(solver, abstraction, next);
        abstraction->result = bad_result;
    }
    assert(abstraction->num_started > 0);
    abstraction->num_started--;
    if (abstraction->num_started == 0) {
        semaphore_post(&abstraction->sub_finished);
    }
    pthread_mutex_unlock(&abstraction->mutex);
    return sub_result;
}

static void* solve_sub_untyped(void* data) {
    ClausalAbstraction* abs = data;
    Solver* solver = abs->solver;
    while (true) {
        semaphore_wait(&abs->has_entry);
        semaphore_wait(&solver->num_threads);
        solve_sub(abs);
        semaphore_post(&solver->num_threads);
    }
    return NULL;
}

static void solve_sub_concurrently(Solver* solver, ClausalAbstraction* abstraction) {
    assert(abstraction->num_started == 0);
    abstraction->num_started++;
    
    assert(solver);
    
    for (size_t i = 0; i < abstraction->scope->num_next; i++) {
        ClausalAbstraction* next = abstraction->next[i];
        assert(next != NULL);
        
        pthread_mutex_lock(&abstraction->mutex);
        abstraction->num_started++;
        pthread_mutex_unlock(&abstraction->mutex);
        if (abstraction->scope->num_next > 1) {
            semaphore_post(&next->has_entry);
        } else {
            solve_sub(next);
        }
    }
    
    pthread_mutex_lock(&abstraction->mutex);
    abstraction->num_started--;
    if (abstraction->num_started == 0) {
        semaphore_post(&abstraction->sub_finished);
    }
    pthread_mutex_unlock(&abstraction->mutex);
    if (abstraction->scope->num_next > 1) {
        semaphore_wait(&abstraction->sub_finished);
    }
}
#else
static void solve_sub_sequentially(Solver* solver, ClausalAbstraction* abstraction) {
    const bool is_existential = abstraction->scope->qtype == QUANTIFIER_EXISTENTIAL;
    const qbf_res good_result = is_existential ? QBF_RESULT_SAT : QBF_RESULT_UNSAT;
    const qbf_res bad_result = is_existential ? QBF_RESULT_UNSAT : QBF_RESULT_SAT;
    
    for (size_t i = 0; i < abstraction->scope->num_next; i++) {
        ClausalAbstraction* next = abstraction->next[i];
        assert(next != NULL);
        
        qbf_res sub_result = solve_recursive(solver, next);
        if (sub_result == good_result) {
            append_to_min_entry(solver, abstraction, next);
        } else {
            assert(sub_result == bad_result);
            refine(solver, abstraction, next);
            abstraction->result = bad_result;
        }
    }
}
#endif


static qbf_res solve_recursive(Solver* solver, ClausalAbstraction* abstraction) {
    const bool is_existential = abstraction->scope->qtype == QUANTIFIER_EXISTENTIAL;
    const qbf_res good_result = is_existential ? QBF_RESULT_SAT : QBF_RESULT_UNSAT;
    const qbf_res bad_result = is_existential ? QBF_RESULT_UNSAT : QBF_RESULT_SAT;
    
    while (true) {
        logging_debug("\n");
        logging_info("%s level %d\n", is_existential ? "existential" : "universal", abstraction->scope->scope_id);
        statistics_start_timer(abstraction->statistics);
        
        assume_t_literals(solver, abstraction, false);
        
        sat_res result = satsolver_sat(abstraction->sat);
        if (result == SATSOLVER_SATISFIABLE) {
            get_assignment(solver, abstraction);
            if (abstraction->scope->num_next == 0) {
                if (solver->options->dual_propagation_minimization) {
                    dual_propagation(solver, abstraction);
                }
                statistics_stop_and_record_timer(abstraction->statistics);
                return good_result;
            }
            assert(abstraction->scope->num_next > 0);
            
            get_assumptions(solver, abstraction);
            
            statistics_stop_and_record_timer(abstraction->statistics);
            
            abstraction->result = good_result;
            int_vector_reset(abstraction->min_entry);
            
#ifdef PARALLEL_SOLVING
            solve_sub_concurrently(solver, abstraction);
#else
            solve_sub_sequentially(solver, abstraction);
#endif
            
            if (abstraction->result == good_result) {
                get_sat_assumptions(solver, abstraction);
                if (solver->options->dual_propagation_minimization) {
                    dual_propagation(solver, abstraction);
                } else {
                    unsat_propagation(solver, abstraction);
                }
                return good_result;
            }
        } else {
            assert(result == SATSOLVER_UNSATISFIABLE);
            get_unsat_core(solver, abstraction);
            statistics_stop_and_record_timer(abstraction->statistics);
            return bad_result;
        }
    }
}

static qbf_res solve(Solver* solver) {
    assert(solver->matrix->top_level->qtype == QUANTIFIER_EXISTENTIAL);
    return solve_recursive(solver, solver->abstraction);
}


/* Interface implementation */

SolverOptions* solver_get_default_options() {
    SolverOptions* default_options = malloc(sizeof(SolverOptions));
    
    default_options->preprocess = true;
    default_options->miniscoping = true;
    default_options->certify = false;
    default_options->statistics = false;
    
    default_options->dual_propagation_minimization = true;
    default_options->strong_unsat_refinement = true;
    default_options->assume_partial_results = true;
    
#ifdef PARALLEL_SOLVING
    default_options->num_threads = 2;
#endif
    
    return default_options;
}

Solver* solver_init(Matrix* matrix, SolverOptions* options) {
    Solver* solver = malloc(sizeof(Solver));
    solver->options = options;
    solver->matrix = matrix;
    solver->abstraction = NULL;
    
#ifdef PARALLEL_SOLVING
    semaphore_init(&solver->num_threads, options->num_threads);
#endif
    
#ifdef CERTIFICATION
    if (options->certify) {
        certification_init(&solver->certificate, matrix);
    }
#endif
    
    // Statistics
    solver->preprocessing = statistics_init(10000);
    solver->building_abstraction = statistics_init(10000);
    solver->solving = statistics_init(10000);
    
    return solver;
}

void solver_free(Solver* solver) {
    matrix_free(solver->matrix);
    free(solver->options);
    // FIXME: free clausal abstractions
    free(solver);
}

qbf_res solver_sat(Solver* solver) {
    statistics_start_timer(solver->preprocessing);
    if (solver->options->preprocess) {
        matrix_preprocess(solver->matrix);
        //matrix_compress(solver->matrix);
    }
    if (solver->matrix->conflict) {
#ifdef CERTIFICATION
        if (solver->options->certify) {
            certification_import_variables(&solver->certificate, solver->matrix);
        }
#endif
        return QBF_RESULT_UNSAT;
    }
    if (solver->options->miniscoping) {
        matrix_unprenex_by_miniscoping(solver->matrix);
    }
    statistics_stop_and_record_timer(solver->preprocessing);
    
    if (logging_get_verbosity() >= VERBOSITY_ALL) {
        matrix_print_qdimacs(solver->matrix);
    }
    
#ifdef CERTIFICATION
    if (solver->options->certify) {
        certification_import_variables(&solver->certificate, solver->matrix);
    }
#endif
    
    statistics_start_timer(solver->building_abstraction);
    solver->abstraction = abstraction_init(solver, solver->matrix->top_level, NULL);
    statistics_stop_and_record_timer(solver->building_abstraction);
    
    statistics_start_timer(solver->solving);
    qbf_res result = solve(solver);
    statistics_stop_and_record_timer(solver->solving);
    
#ifdef CERTIFICATION
    if (solver->options->certify) {
        certification_define_outputs(&solver->certificate, solver->abstraction, result);
    }
#endif
    
    return result;
}


static void print_scope_statistics_recursively(ClausalAbstraction* abs) {
    printf("statistics for %s level %d\n", abs->scope->qtype == QUANTIFIER_EXISTENTIAL ? "existential" : "universal", abs->scope->scope_id);
    statistics_print(abs->statistics);
    
    for (size_t i = 0; i < abs->scope->num_next; i++) {
        print_scope_statistics_recursively(abs->next[i]);
    }
}

void solver_print_statistics(Solver* solver) {
    printf("preprocessing took ");
    statistics_print_time(solver->preprocessing);
    
    printf("building abstraction took ");
    statistics_print_time(solver->building_abstraction);
    
    printf("solving took ");
    statistics_print_time(solver->solving);
    
    printf("\ndetailed solving statistics:\n");
    if (solver->abstraction != NULL) {
        print_scope_statistics_recursively(solver->abstraction);
    }
}

const char* solver_get_version() {
    return VERSION_STRING;
}
